import os
import sys
from sentence_transformers.readers import InputExample
import logging

logger = logging.getLogger(__name__)

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

    


def load_data_as_individuals(cluster_dict, type,already_clustered=False):
    '''
    This is for SupCon Loss

    '''
  
    
    indv_data = []
    guid = 1
    clus_id=1
    for cluster_id in list(cluster_dict.keys()):
        

        for text in cluster_dict[cluster_id]:
            indv_data.append(InputExample(guid=guid, texts=[text], label=clus_id))
            guid += 1
        clus_id +=1

    print(f'{len(indv_data)} {type} examples')
    return indv_data




import os
import sys
import numpy as np
import pandas as pd
from typing import List
from sentence_transformers import evaluation, LoggingHandler
import logging
from transformers import logging as lg
import wandb
from typing import List, Dict, Set, Callable
from torch import Tensor

from sklearn.metrics.pairwise import paired_cosine_distances, paired_euclidean_distances, paired_manhattan_distances
from sklearn.metrics import average_precision_score

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

from tqdm.autonotebook import trange
import torch
import heapq

lg.set_verbosity_error()
logging.basicConfig(format='%(asctime)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    level=logging.INFO,
                    handlers=[LoggingHandler()])
logger = logging.getLogger(__name__)



def cos_sim(a: Tensor, b: Tensor):
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.
    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    a_norm = torch.nn.functional.normalize(a, p=2, dim=1)
    b_norm = torch.nn.functional.normalize(b, p=2, dim=1)
    return torch.mm(a_norm, b_norm.transpose(0, 1))


def dot_score(a: Tensor, b: Tensor):
    """
    Computes the dot-product dot_prod(a[i], b[j]) for all i and j.
    :return: Matrix with res[i][j]  = dot_prod(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    return torch.mm(a, b.transpose(0, 1))


class InformationRetrievalEvaluator_wandb(evaluation.InformationRetrievalEvaluator):
    """Information Retrieval """
    def __init__(self,
                 queries: Dict[str, str],  #qid => query
                 corpus: Dict[str, str],  #cid => doc
                 relevant_docs: Dict[str, Set[str]],  #qid => Set[cid]
                 corpus_chunk_size: int = 50000,
                 mrr_at_k: List[int] = [10],
                 ndcg_at_k: List[int] = [10],
                 accuracy_at_k: List[int] = [1, 3, 5, 10],
                 precision_recall_at_k: List[int] = [1, 3, 5, 10],
                 map_at_k: List[int] = [100],
                 show_progress_bar: bool = False,
                 batch_size: int = 32,
                 name: str = '',
                 write_csv: bool = True,
                 score_functions: List[Callable[[Tensor, Tensor], Tensor] ] = {'cos_sim': cos_sim},       #Score function, higher=more similar
                 main_score_function: str = None,
                 wandb_names=None
                 ):

        self.queries_ids = []
        for qid in queries:
            if qid in relevant_docs and len(relevant_docs[qid]) > 0:
                self.queries_ids.append(qid)
        self.queries_raw= queries
        self.corpus_raw= corpus

        self.queries = [queries[qid] for qid in self.queries_ids]

        self.corpus_ids = list(corpus.keys())
        self.corpus = [corpus[cid] for cid in self.corpus_ids]

        self.relevant_docs = relevant_docs
        self.corpus_chunk_size = corpus_chunk_size
        self.mrr_at_k = mrr_at_k
        self.ndcg_at_k = ndcg_at_k
        self.accuracy_at_k = accuracy_at_k
        self.precision_recall_at_k = precision_recall_at_k
        self.map_at_k = map_at_k

        self.show_progress_bar = show_progress_bar
        self.batch_size = batch_size
        self.name = name
        self.write_csv = write_csv
        self.score_functions = score_functions
        self.score_function_names = sorted(list(self.score_functions.keys()))
        self.main_score_function = main_score_function
        self.wandb_names = wandb_names

        if name:
            name = "_" + name

        self.csv_file: str = "Information-Retrieval_evaluation" + name + "_results.csv"
        self.csv_headers = ["epoch", "steps"]

        for score_name in self.score_function_names:
            for k in accuracy_at_k:
                self.csv_headers.append("{}-Accuracy@{}".format(score_name, k))

            for k in precision_recall_at_k:
                self.csv_headers.append("{}-Precision@{}".format(score_name, k))
                self.csv_headers.append("{}-Recall@{}".format(score_name, k))

            for k in mrr_at_k:
                self.csv_headers.append("{}-MRR@{}".format(score_name, k))

            for k in ndcg_at_k:
                self.csv_headers.append("{}-NDCG@{}".format(score_name, k))

            for k in map_at_k:
                self.csv_headers.append("{}-MAP@{}".format(score_name, k))


    def __call__(self, model, output_path: str = None, epoch: int = -1, steps: int = -1, *args, **kwargs) -> float:
        if epoch != -1:
            out_txt = " after epoch {}:".format(epoch) if steps == -1 else " in epoch {} after {} steps:".format(epoch, steps)
        else:
            out_txt = ":"

        logger.info("Information Retrieval Evaluation on " + self.name + " dataset" + out_txt)

        scores = self.compute_metrices(model, *args, **kwargs)

        ##Log the scores
        for key in scores:
            logger.info(key + ": " + str(scores[key]))
            if self.wandb_names is not None:
                wandb.log({key: scores[key]})


        # Write results to disc
        if output_path is not None and self.write_csv:
            csv_path = os.path.join(output_path, self.csv_file)
            if not os.path.isfile(csv_path):
                fOut = open(csv_path, mode="w", encoding="utf-8")
                fOut.write(",".join(self.csv_headers))
                fOut.write("\n")

            else:
                fOut = open(csv_path, mode="a", encoding="utf-8")

            output_data = [epoch, steps]
            for name in self.score_function_names:
                for k in self.accuracy_at_k:
                    output_data.append(scores[name]['accuracy@k'][k])

                for k in self.precision_recall_at_k:
                    output_data.append(scores[name]['precision@k'][k])
                    output_data.append(scores[name]['recall@k'][k])

                for k in self.mrr_at_k:
                    output_data.append(scores[name]['mrr@k'][k])

                for k in self.ndcg_at_k:
                    output_data.append(scores[name]['ndcg@k'][k])

                for k in self.map_at_k:
                    output_data.append(scores[name]['map@k'][k])

            fOut.write(",".join(map(str, output_data)))
            fOut.write("\n")
            fOut.close()

        

        if self.main_score_function is None:
            print("Using max accuracy@k as main score - ", scores[name]['accuracy@k'][1])
            return max([scores[name]['accuracy@k'][1] for name in self.score_function_names])
        
        else:
            print("Using max accuracy@k as main score - ", scores[name]['accuracy@k'][1])
            return scores[self.main_score_function]['accuracy@k'][1]    

    def calculate_matching_accuracy(self, queries_result_list) -> float:
        matches = []
        for query_itr in range(len(self.queries)):
            query = self.queries[query_itr]
            relevant_docs = self.relevant_docs_dic[query]
            top_k_results = queries_result_list[query_itr]

            for doc in top_k_results:
                corpus_id = doc['corpus_id']
                if corpus_id in relevant_docs:
                    match = {"query": query, "match": corpus_id, "score": doc['score']}
                    matches.append(match)
                    break

        matching_accuracy = len(matches) / len(self.queries)

        # Save the matching results to a CSV file
        matching_df = pd.DataFrame(matches)
        matching_df.to_csv("matching_output.csv", index=False)

        return matching_accuracy



    def compute_metrices(self, model, corpus_model = None, corpus_embeddings: Tensor = None) -> Dict[str, float]:
        if corpus_model is None:
            corpus_model = model

        max_k = max(max(self.mrr_at_k), max(self.ndcg_at_k), max(self.accuracy_at_k), max(self.precision_recall_at_k), max(self.map_at_k))

        # Compute embedding for the queries
        query_embeddings = model.encode(self.queries, show_progress_bar=self.show_progress_bar, batch_size=self.batch_size, convert_to_tensor=True,normalize_embeddings=True)

        queries_result_list = {}
        for name in self.score_functions:
            queries_result_list[name] = [[] for _ in range(len(query_embeddings))]

        #Iterate over chunks of the corpus
        for corpus_start_idx in trange(0, len(self.corpus), self.corpus_chunk_size, desc='Corpus Chunks', disable=not self.show_progress_bar):
            corpus_end_idx = min(corpus_start_idx + self.corpus_chunk_size, len(self.corpus))

            #Encode chunk of corpus
            if corpus_embeddings is None:
                sub_corpus_embeddings = corpus_model.encode(self.corpus[corpus_start_idx:corpus_end_idx], show_progress_bar=False, batch_size=self.batch_size, convert_to_tensor=True)
            else:
                sub_corpus_embeddings = corpus_embeddings[corpus_start_idx:corpus_end_idx]

            #Compute cosine similarites
            for name, score_function in self.score_functions.items():
                pair_scores = score_function(query_embeddings, sub_corpus_embeddings)

                #Get top-k values
                pair_scores_top_k_values, pair_scores_top_k_idx = torch.topk(pair_scores, min(max_k, len(pair_scores[0])), dim=1, largest=True, sorted=False)
                pair_scores_top_k_values = pair_scores_top_k_values.cpu().tolist()
                pair_scores_top_k_idx = pair_scores_top_k_idx.cpu().tolist()

                for query_itr in range(len(query_embeddings)):
                    for sub_corpus_id, score in zip(pair_scores_top_k_idx[query_itr], pair_scores_top_k_values[query_itr]):
                        corpus_id = self.corpus_ids[corpus_start_idx+sub_corpus_id]
                        if len(queries_result_list[name][query_itr]) < max_k:
                            heapq.heappush(queries_result_list[name][query_itr], (score, corpus_id))  # heaqp tracks the quantity of the first element in the tuple
                        else:
                            heapq.heappushpop(queries_result_list[name][query_itr], (score, corpus_id))
        for name in queries_result_list:
            for query_itr in range(len(queries_result_list[name])):
                for doc_itr in range(len(queries_result_list[name][query_itr])):
                    score, corpus_id = queries_result_list[name][query_itr][doc_itr]
                    queries_result_list[name][query_itr][doc_itr] = {'corpus_id': corpus_id, 'score': score}
        logger.info("Queries: {}".format(len(self.queries)))
        logger.info("Corpus: {}\n".format(len(self.corpus)))

        #Compute scores
        scores = {name: self.compute_metrics(queries_result_list[name]) for name in self.score_functions}

        #Output
        for name in self.score_function_names:
            logger.info("Score-Function: {}".format(name))
            self.output_scores(scores[name])

        return scores


    def compute_metrics(self, queries_result_list: List[object],save_best_matches=True):
        # Init score computation values
        num_hits_at_k = {k: 0 for k in self.accuracy_at_k}
        precisions_at_k = {k: [] for k in self.precision_recall_at_k}
        recall_at_k = {k: [] for k in self.precision_recall_at_k}
        MRR = {k: 0 for k in self.mrr_at_k}
        ndcg = {k: [] for k in self.ndcg_at_k}
        AveP_at_k = {k: [] for k in self.map_at_k}

        best_match_dict={k_matches:{} for k_matches in self.accuracy_at_k}


        # Compute scores on results
        for query_itr in range(len(queries_result_list)):
            query_id = self.queries_ids[query_itr]
            query_text= self.queries_raw[query_id]

            # Sort scores
            top_hits = sorted(queries_result_list[query_itr], key=lambda x: x['score'], reverse=True)
            query_relevant_docs = self.relevant_docs[query_id]

            # Accuracy@k - We count the result correct, if at least one relevant doc is accross the top-k documents
            for k_val in self.accuracy_at_k:
                for hit in top_hits[0:k_val]:
                    if hit['corpus_id'] in query_relevant_docs:
                        num_hits_at_k[k_val] += 1
                        break
            #Fill up best_match_dict
            for k_val in self.accuracy_at_k:
                ##Make a tuple with correct match or not (1/0) and the hit
                for hit in top_hits[0:k_val]:
                    if hit['corpus_id'] in query_relevant_docs:
                        best_match_dict[k_val][query_text]=(1,self.corpus_raw[hit['corpus_id']])
                        break
                    else:
                        best_match_dict[k_val][query_text]=(0,self.corpus_raw[hit['corpus_id']])
            

            # Precision and Recall@k
            for k_val in self.precision_recall_at_k:
                num_correct = 0
                for hit in top_hits[0:k_val]:
                    if hit['corpus_id'] in query_relevant_docs:
                        num_correct += 1

                precisions_at_k[k_val].append(num_correct / k_val)
                recall_at_k[k_val].append(num_correct / len(query_relevant_docs))

            # MRR@k
            for k_val in self.mrr_at_k:
                for rank, hit in enumerate(top_hits[0:k_val]):
                    if hit['corpus_id'] in query_relevant_docs:
                        MRR[k_val] += 1.0 / (rank + 1)
                        break

            # NDCG@k
            for k_val in self.ndcg_at_k:
                predicted_relevance = [1 if top_hit['corpus_id'] in query_relevant_docs else 0 for top_hit in top_hits[0:k_val]]
                true_relevances = [1] * len(query_relevant_docs)

                ndcg_value = self.compute_dcg_at_k(predicted_relevance, k_val) / self.compute_dcg_at_k(true_relevances, k_val)
                ndcg[k_val].append(ndcg_value)

            # MAP@k
            for k_val in self.map_at_k:
                num_correct = 0
                sum_precisions = 0

                for rank, hit in enumerate(top_hits[0:k_val]):
                    if hit['corpus_id'] in query_relevant_docs:
                        num_correct += 1
                        sum_precisions += num_correct / (rank + 1)

                avg_precision = sum_precisions / min(k_val, len(query_relevant_docs))
                AveP_at_k[k_val].append(avg_precision)

        # Compute averages
        for k in num_hits_at_k:
            num_hits_at_k[k] /= len(self.queries)

        for k in precisions_at_k:
            precisions_at_k[k] = np.mean(precisions_at_k[k])

        for k in recall_at_k:
            recall_at_k[k] = np.mean(recall_at_k[k])

        for k in ndcg:
            ndcg[k] = np.mean(ndcg[k])

        for k in MRR:
            MRR[k] /= len(self.queries)

        for k in AveP_at_k:
            AveP_at_k[k] = np.mean(AveP_at_k[k])

        ##Save the best matches as a json
        if save_best_matches:
            # with open('best_matches.json', 'w') as fp:
            #     json.dump(best_match_dict, fp)
            
            ##Make a df with the matches. keys, value[0], value[1] as the three columns
            df = pd.DataFrame(list(best_match_dict[1].items()), columns=["Query", "Match"])
            df["Correct"] = df["Match"].apply(lambda x: x[1])
            df["Match"] = df["Match"].apply(lambda x: x[0])

            ###save the df as a csv 
            df.to_csv("current_best_matches.csv", index=False,encoding="utf-8")

 
        return {'accuracy@k': num_hits_at_k, 'precision@k': precisions_at_k, 'recall@k': recall_at_k, 'ndcg@k': ndcg, 'mrr@k': MRR, 'map@k': AveP_at_k}




class BinaryClassificationEvaluator_wandb(evaluation.BinaryClassificationEvaluator):


    def __init__(self, sentences1: List[str], sentences2: List[str], labels: List[int], name: str = '', batch_size: int = 32, show_progress_bar: bool = False, write_csv: bool = True, wandb_names=None):
        self.sentences1 = sentences1
        self.sentences2 = sentences2
        self.labels = labels
        self.wandb_names = wandb_names

        assert len(self.sentences1) == len(self.sentences2)
        assert len(self.sentences1) == len(self.labels)
        for label in labels:
            assert (label == 0 or label == 1)

        self.write_csv = write_csv
        self.name = name
        self.batch_size = batch_size
        if show_progress_bar is None:
            show_progress_bar = (logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG)
        self.show_progress_bar = show_progress_bar

        self.csv_file = "binary_classification_evaluation" + ("_"+name if name else '') + "_results.csv"
        self.csv_headers = ["epoch", "steps",
                            "cossim_accuracy", "cossim_accuracy_threshold", "cossim_f1", "cossim_precision", "cossim_recall", "cossim_f1_threshold", "cossim_ap",
                            "manhattan_accuracy", "manhattan_accuracy_threshold", "manhattan_f1", "manhattan_precision", "manhattan_recall", "manhattan_f1_threshold", "manhattan_ap",
                            "euclidean_accuracy", "euclidean_accuracy_threshold", "euclidean_f1", "euclidean_precision", "euclidean_recall", "euclidean_f1_threshold", "euclidean_ap",
                            "dot_accuracy", "dot_accuracy_threshold", "dot_f1", "dot_precision", "dot_recall", "dot_f1_threshold", "dot_ap"]



    def compute_metrices(self, model):

        print("*********************EVALUATION STARTED******************************")

        sentences = list(set(self.sentences1 + self.sentences2))
        embeddings = model.encode(sentences, batch_size=self.batch_size, show_progress_bar=self.show_progress_bar, convert_to_numpy=True)
        emb_dict = {sent: emb for sent, emb in zip(sentences, embeddings)}
        embeddings1 = [emb_dict[sent] for sent in self.sentences1]
        embeddings2 = [emb_dict[sent] for sent in self.sentences2]

        cosine_scores = 1 - paired_cosine_distances(embeddings1, embeddings2)
        manhattan_distances = paired_manhattan_distances(embeddings1, embeddings2)
        euclidean_distances = paired_euclidean_distances(embeddings1, embeddings2)

        embeddings1_np = np.asarray(embeddings1)
        embeddings2_np = np.asarray(embeddings2)
        dot_scores = [np.dot(embeddings1_np[i], embeddings2_np[i]) for i in range(len(embeddings1_np))]

        labels = np.asarray(self.labels)
        output_scores = {}
        for short_name, name, scores, reverse in [['cossim', 'Cosine-Similarity', cosine_scores, True], ['manhattan', 'Manhattan-Distance', manhattan_distances, False], ['euclidean', 'Euclidean-Distance', euclidean_distances, False], ['dot', 'Dot-Product', dot_scores, True]]:
            # Note: newer versions of sbert have updated the spelling on manhatten to manhattan

            acc, acc_threshold = self.find_best_acc_and_threshold(scores, labels, reverse)
            f1, precision, recall, f1_threshold = self.find_best_f1_and_threshold(scores, labels, reverse)
            ap = average_precision_score(labels, scores * (1 if reverse else -1))

            logger.info("Accuracy with {}:           {:.2f}\t(Threshold: {:.4f})".format(name, acc * 100, acc_threshold))
            logger.info("F1 with {}:                 {:.2f}\t(Threshold: {:.4f})".format(name, f1 * 100, f1_threshold))
            logger.info("Precision with {}:          {:.2f}".format(name, precision * 100))
            logger.info("Recall with {}:             {:.2f}".format(name, recall * 100))
            logger.info("Average Precision with {}:  {:.2f}\n".format(name, ap * 100))

            output_scores[short_name] = {
                'accuracy': acc,
                'accuracy_threshold': acc_threshold,
                'f1': f1,
                'f1_threshold': f1_threshold,
                'precision': precision,
                'recall': recall,
                'ap': ap
            }
            if self.wandb_names:
                wandb.log({
                    f"Classification Accuracy {name}": acc,
                    f"Classification Accuracy threshold {name}": acc_threshold,
                    f"Classification F1 {name}": f1,
                    f"Classification F1 threshold {name}": f1_threshold,
                    f"Classification Precision {name}": precision,
                    f"Classification Recall {name}": recall,
                    f"Classification Average precision {name}": ap
                })

        return output_scores


# __init__.py

__version__ = "0.1.8"
__MODEL_HUB_ORGANIZATION__ = 'sentence-transformers' #For compatibility with sentence-transformers
from .data import DATA_DIR_PATH
from .infer import *
from .preprocess import *
from .train_model import *
from .modified_sbert import *
from .train_clf_model import train_clf_model
from .modelling.LinkTransformer import LinkTransformer
from .modelling.LinkTransformerClassifier import LinkTransformerClassifier





from itertools import combinations
from sklearn.cluster import AgglomerativeClustering, DBSCAN
import networkx as nx
import hdbscan


def cluster(cluster_type, cluster_params, corpus_embeddings, corpus_ids=None):
    # Define default values for each cluster type
    default_params = {
        "agglomerative": {
            "threshold": 0.5,
            "clustering linkage": "ward",  # You can choose a default linkage method
            "metric": "euclidean",  # You can choose a default metric
        },
        "HDBScan": {
            "min cluster size": 5,
            "min samples": 1,
        },
        "SLINK": {
            "min cluster size": 2,
            "threshold": 0.1,
        },
    }

    if cluster_type not in ["agglomerative", "HDBScan", "SLINK"]:
        raise ValueError('cluster_type must be "agglomerative", "HDBScan", or "SLINK"')

    # Get the default parameters based on cluster_type
    default_params_for_type = default_params.get(cluster_type, {})

    # Update cluster_params with default values for missing keys
    for key, default_value in default_params_for_type.items():
        cluster_params.setdefault(key, default_value)

    # Validate the cluster_params
    if cluster_type == "agglomerative":
        required_params = ["threshold", "clustering linkage", "metric"]
    elif cluster_type == "HDBScan":
        required_params = ["min cluster size", "min samples"]
    elif cluster_type == "SLINK":
        required_params = ["min cluster size", "threshold"]
    else:
        raise ValueError('Invalid cluster_type')

    for param in required_params:
        if param not in cluster_params:
            raise ValueError(f'cluster_params must contain "{param}"')

    if cluster_type == "agglomerative":
        clustering_model = AgglomerativeClustering(
            n_clusters=None,
            distance_threshold=cluster_params["threshold"],
            linkage=cluster_params["clustering linkage"],
            metric=cluster_params["metric"]
        )

    if cluster_type == "SLINK":
        clustering_model = DBSCAN(
            eps=cluster_params["threshold"],
            min_samples=cluster_params["min cluster size"],
            metric=cluster_params["metric"]
        )

    if cluster_type == "HDBScan":
        clustering_model = hdbscan.HDBSCAN(
            min_cluster_size=cluster_params["min cluster size"],
            min_samples=cluster_params["min samples"],
            gen_min_span_tree=True
        )

    clustering_model.fit(corpus_embeddings)
    cluster_assignment = clustering_model.labels_


    return cluster_assignment


def clusters_from_edges(edges_list):
    """Identify clusters of passages given a dictionary of edges"""

    # clusters via NetworkX
    G = nx.Graph()
    G.add_edges_from(edges_list)
    sub_graphs = [G.subgraph(c).copy() for c in nx.connected_components(G)]

    sub_graph_dict = {}
    for i in range(len(sub_graphs)):
        sub_graph_dict[i] = list(sub_graphs[i].nodes())

    return sub_graph_dict


def edges_from_clusters(cluster_dict):
    """
    Convert every pair in a cluster into an edge
    """
    cluster_edges = []
    for cluster_id in list(cluster_dict.keys()):
        art_ids_list = cluster_dict[cluster_id]
        edge_list = [list(comb) for comb in combinations(art_ids_list, 2)]
        cluster_edges.extend(edge_list)

    return cluster_edges



from transformers import AutoModelForSequenceClassification, AutoTokenizer
from typing import Optional, Union, Iterable, List, Dict, Tuple, Callable, Any
import os
import logging
from huggingface_hub import HfApi, HfFolder, Repository
from sentence_transformers import __version__ 
import os
import json
import shutil
import stat
from linktransformer import __MODEL_HUB_ORGANIZATION__
import tempfile
from distutils.dir_util import copy_tree

from linktransformer.model_card_templates import ClassificationModelCardTemplate

logger = logging.getLogger(__name__)

##Define methods for 1) Saving model card with LT details, 2)Upload to huggingface (model + tokenizer)
class LinkTransformerClassifier:
    """Modified Sequence Classification model and tokenizer to implement model card generation and save to hub functions """
    def __init__(self, model_name_or_path: str,
                 opt_model_description: Optional[str] = None,
                 opt_model_lang: Optional[str] = None,
                 label_map=None,model_card_text=None):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        self.model = AutoModelForSequenceClassification.from_pretrained(model_name_or_path)
        self.label_map = label_map
        self.model_name_or_path = model_name_or_path
        self.opt_model_description = opt_model_description
        self.opt_model_lang = opt_model_lang
        self._model_card_text = model_card_text
    

        ###Find the config file and get the base model name
        config_path = os.path.join(model_name_or_path, 'config.json')
        if os.path.exists(config_path):
            with open(config_path, 'r') as fIn:
                config = json.load(fIn)
                self.base_model_name_or_path = config['model_type']
        else:
            self.base_model_name_or_path = model_name_or_path



        ##If label map is none, try and find it in the model dir
        if self.label_map is None:
            label_map_path = os.path.join(model_name_or_path, 'label_map.json')
            if os.path.exists(label_map_path):
                with open(label_map_path, 'r') as fIn:
                    self.label_map = json.load(fIn)
        
    def save(self, save_directory: str,model_name: Optional[str] = None,override_model_description: Optional[str] = None, override_model_lang: Optional[str] = None,train_datasets=None):
        """
        Saves the model and tokenizer to the specified directory.
        """
        if override_model_description is not None:
            self.opt_model_description = override_model_description
        if override_model_lang is not None:
            self.opt_model_lang = override_model_lang
        self.tokenizer.save_pretrained(save_directory)
        self.model.save_pretrained(save_directory)
        self._create_model_card(save_directory, model_name=model_name,label_map=self.label_map,train_datasets=train_datasets)
        
    def _create_model_card(self, path: str, model_name: Optional[str] = None,label_map=None,train_datasets: Optional[List[str]] = None):
            """
            Create an automatic model and stores it in path
            """
            if self._model_card_text is not None and len(self._model_card_text) > 0:
                model_card = self._model_card_text
            else:
                tags = ClassificationModelCardTemplate.__TAGS__.copy()
                model_card = ClassificationModelCardTemplate.__MODEL_CARD__

                # Print full model
                model_card = model_card.replace("{FULL_MODEL_STR}", str(self))

                # Add tags
                model_card = model_card.replace("{TAGS}", "\n".join(["- "+t for t in tags]))

                datasets_str = ""
                if train_datasets is not None:
                    datasets_str = "datasets:\n"+"\n".join(["- " + d for d in train_datasets])
                model_card = model_card.replace("{DATASETS}", datasets_str)


                ###Add base model name
                model_card=model_card.replace("{BASE_MODEL}",self.base_model_name_or_path)

                ##Add label map
                if label_map is not None:
                    label_map_str = ""
                    for key, value in label_map.items():
                        label_map_str += f"- {key}: {value}\n"
                    model_card = model_card.replace("{LABEL_MAP}", label_map_str)
                

                ##ADd optional model description
                if self.opt_model_description is not None:
                    model_card=model_card.replace("{MODEL_DESCRIPTION}",self.opt_model_description)
                else:
                    model_card=model_card.replace("{MODEL_DESCRIPTION}","")
                    
                ##ADd optional model language
                if self.opt_model_lang is not None:
                    if isinstance(self.opt_model_lang,list):
                        model_card = model_card.replace("{LANGUAGE}", "\n".join(["- "+t for t in self.opt_model_lang]))
                    else:
                        model_card = model_card.replace("{LANGUAGE}", "\n".join(["- "+t for t in [self.opt_model_lang]]))
                else:
                    model_card = model_card.replace("language: \n{LANGUAGE}", "")
                    model_card = model_card.replace("It is pretrained for the language: {LANGUAGE}.", "")



                # Replace remaining vars with default values
                for name, value in ClassificationModelCardTemplate.__DEFAULT_VARS__.items():
                    model_card = model_card.replace(name, str(value))

            if model_name is not None:
                model_card = model_card.replace("{MODEL_NAME}", model_name.strip())

            with open(os.path.join(path, "README.md"), "w", encoding='utf8') as fOut:
                fOut.write(model_card.strip())


    def save_to_hub(self,
                    repo_name: str,
                    organization: Optional[str] = None,
                    private: Optional[bool] = None,
                    commit_message: str = "Add new LinkTransformer model.",
                    local_model_path: Optional[str] = None,
                    exist_ok: bool = False,
                    override_model_description: Optional[str] = None, 
                    override_model_lang: Optional[str] = None,
                    train_datasets: Optional[List[str]] = None):
        """
        Uploads all elements of this LinkTransformer (for classification) to a new HuggingFace Hub repository.

        :param repo_name: Repository name for your model in the Hub.
        :param organization:  Organization in which you want to push your model or tokenizer (you must be a member of this organization).
        :param private: Set to true, for hosting a prive model
        :param commit_message: Message to commit while pushing.
        :param local_model_path: Path of the model locally. If set, this file path will be uploaded. Otherwise, the current model will be uploaded
        :param exist_ok: If true, saving to an existing repository is OK. If false, saving only to a new repository is possible
        :param replace_model_card: If true, replace an existing model card in the hub with the automatically created model card
        :param train_datasets: Datasets used to train the model. If set, the datasets will be added to the model card in the Hub.
        :return: The url of the commit of your model in the given repository.
        """
        token = HfFolder.get_token()
        if token is None:
            raise ValueError("You must login to the Hugging Face hub on this computer by typing `transformers-cli login`.")

        if '/' in repo_name:
            splits = repo_name.split('/', maxsplit=1)
            if organization is None or organization == splits[0]:
                organization = splits[0]
                repo_name = splits[1]
            else:
                raise ValueError("You passed and invalid repository name: {}.".format(repo_name))

        endpoint = "https://huggingface.co"
        repo_id = repo_name
        if organization:
            repo_id = f"{organization}/{repo_id}"
        repo_url = HfApi(endpoint=endpoint).create_repo(
                repo_id=repo_id,
                token=token,
                private=private,
                repo_type=None,
                exist_ok=exist_ok,
            )
        full_model_name = repo_url[len(endpoint)+1:].strip("/")

        with tempfile.TemporaryDirectory() as tmp_dir:
            # First create the repo (and clone its content if it's nonempty).
            logger.info("Create repository and clone it if it exists")
            repo = Repository(tmp_dir, clone_from=repo_url)

            # If user provides local files, copy them.
            if local_model_path:
                copy_tree(local_model_path, tmp_dir)
            else:  # Else, save model directly into local repo.
                self.save(tmp_dir, full_model_name,override_model_description, override_model_lang,train_datasets)

            ##Save the model config (_lt_model_config) to the repo if it exists
            if hasattr(self, '_lt_model_config'):
                with open(os.path.join(tmp_dir, 'LT_training_config.json'), 'w') as fOut:
                    json.dump(self._lt_model_config, fOut, indent=2)

            #Find files larger 5M and track with git-lfs
            large_files = []
            for root, dirs, files in os.walk(tmp_dir):
                for filename in files:
                    file_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(file_path, tmp_dir)

                    if os.path.getsize(file_path) > (5 * 1024 * 1024):
                        large_files.append(rel_path)

            if len(large_files) > 0:
                logger.info("Track files with git lfs: {}".format(", ".join(large_files)))
                repo.lfs_track(large_files)

            logger.info("Push model to the hub. This might take a while")
            push_return = repo.push_to_hub(commit_message=commit_message)

            def on_rm_error(func, path, exc_info):
                # path contains the path of the file that couldn't be removed
                # let's just assume that it's read-only and unlink it.
                try:
                    os.chmod(path, stat.S_IWRITE)
                    os.unlink(path)
                except:
                    pass

            # Remove .git folder. On Windows, the .git folder might be read-only and cannot be deleted
            # Hence, try to set write permissions on error
            try:
                for f in os.listdir(tmp_dir):
                    shutil.rmtree(os.path.join(tmp_dir, f), onerror=on_rm_error)
            except Exception as e:
                logger.warning("Error when deleting temp folder: {}".format(str(e)))
                pass


        return push_return





###Modifications to the Base SentenceTransformers class. We need to re-write the saving functions to make LinkTransformer models out of them.
import logging
from sentence_transformers import SentenceTransformer
from typing import Optional, List, Iterable, Union
from huggingface_hub import HfApi, HfFolder, Repository
import transformers
from sentence_transformers import __version__ 
import os
import torch
import json
from linktransformer.model_card_templates import ModelCardTemplate
import shutil
import stat
from sentence_transformers.models import Transformer
from torch import nn
from linktransformer import __MODEL_HUB_ORGANIZATION__
import tempfile
from distutils.dir_util import copy_tree





logger = logging.getLogger(__name__)

class LinkTransformer(SentenceTransformer):
    """
    Modified SentenceTransformer class for LinkTransformers models as a wrapper around the SentenceTransformer class.
    """

    """
    Loads or create a LinkTransformers model, that can be used to map sentences / text to embeddings which in-turn can be used for LM based record linkage.

    :param model_name_or_path: If it is a filepath on disc, it loads the model from that path. If it is not a path, it first tries to download a pre-trained SentenceTransformer model. If that fails, tries to construct a model from Huggingface models repository with that name.
    :param modules: This parameter can be used to create custom SentenceTransformer models from scratch.
    :param device: Device (like 'cuda' / 'cpu') that should be used for computation. If None, checks if a GPU can be used.
    :param cache_folder: Path to store models. Can be also set by SENTENCE_TRANSFORMERS_HOME enviroment variable.
    :param use_auth_token: HuggingFace authentication token to download private models.
    """
    def __init__(self, model_name_or_path: Optional[str] = None,
                 modules: Optional[Iterable[nn.Module]] = None,
                 device: Optional[str] = None,
                 cache_folder: Optional[str] = None,
                 use_auth_token: Union[bool, str, None] = None,
                 opt_model_description: Optional[str] = None,
                 opt_model_lang: Optional[str] = None,
                 ):


        super().__init__(model_name_or_path=model_name_or_path, modules=modules, device=device, cache_folder=cache_folder, use_auth_token=use_auth_token)

        ##If it is a local path, we need to load the config (LT) from the file FLAG - CLEAN THIS UP
        if os.path.isdir(model_name_or_path):
            ###If config file exists, load it. It will be in the parent folder of the model
            if os.path.isfile(os.path.join(model_name_or_path, 'LT_training_config.json')):
                with open(os.path.join(model_name_or_path, 'LT_training_config.json'), 'r') as fIn:
                    self._lt_model_config = json.load(fIn)
                    print("Loaded LinkTransformer model config from {}".format(os.path.join(model_name_or_path, 'LT_training_config.json')))
                self.base_model_name_or_path = self._lt_model_config['base_model_path']
                self.opt_model_description = self._lt_model_config['opt_model_description']
                self.opt_model_lang = self._lt_model_config['opt_model_lang']
        else:
            ###If huggningface model, then assign ##Implement later
            self.base_model_name_or_path = model_name_or_path

            self.opt_model_description = opt_model_description
            self.opt_model_lang = opt_model_lang
            



    def save(self, path: str, model_name: Optional[str] = None, create_model_card: bool = True, train_datasets: Optional[List[str]] = None,override_model_description: Optional[str] = None, 
                    override_model_lang: Optional[str] = None):
        """
        Saves all elements for this seq. sentence embedder into different sub-folders
        :param path: Path on disc
        :param model_name: Optional model name
        :param create_model_card: If True, create a README.md with basic information about this model
        :param train_datasets: Optional list with the names of the datasets used to to train the model
        """
        if path is None:
            return
        
        if override_model_description is not None:
            self.opt_model_description = override_model_description
        if override_model_lang is not None:
            self.opt_model_lang = override_model_lang

        os.makedirs(path, exist_ok=True)

        logger.info("Save model to {}".format(path))
        modules_config = []

        #Save some model info
        if '__version__' not in self._model_config:
            self._model_config['__version__'] = {
                    'sentence_transformers': __version__,
                    'transformers': transformers.__version__,
                    'pytorch': torch.__version__,
                }

        with open(os.path.join(path, 'config_sentence_transformers.json'), 'w') as fOut:
            json.dump(self._model_config, fOut, indent=2)

        #Save modules
        for idx, name in enumerate(self._modules):
            module = self._modules[name]
            if idx == 0 and isinstance(module, Transformer):    #Save transformer model in the main folder
                model_path = path + "/"
            else:
                model_path = os.path.join(path, str(idx)+"_"+type(module).__name__)

            os.makedirs(model_path, exist_ok=True)
            module.save(model_path)
            modules_config.append({'idx': idx, 'name': name, 'path': os.path.basename(model_path), 'type': type(module).__module__})

        with open(os.path.join(path, 'modules.json'), 'w') as fOut:
            json.dump(modules_config, fOut, indent=2)

        # Create model card
        if create_model_card:
            self._create_model_card(path, model_name, train_datasets)

    def _create_model_card(self, path: str, model_name: Optional[str] = None, train_datasets: Optional[List[str]] = None):
        """
        Create an automatic model and stores it in path
        """
        if self._model_card_text is not None and len(self._model_card_text) > 0:
            model_card = self._model_card_text
        else:
            tags = ModelCardTemplate.__TAGS__.copy()
            model_card = ModelCardTemplate.__MODEL_CARD__

            # Print full model
            model_card = model_card.replace("{FULL_MODEL_STR}", str(self))

            # Add tags
            model_card = model_card.replace("{TAGS}", "\n".join(["- "+t for t in tags]))

            datasets_str = ""
            if train_datasets is not None:
                datasets_str = "datasets:\n"+"\n".join(["- " + d for d in train_datasets])
            model_card = model_card.replace("{DATASETS}", datasets_str)

            ###Add base model name
            model_card=model_card.replace("{BASE_MODEL}",self.base_model_name_or_path)

            ##ADd optional model description
            if self.opt_model_description is not None:
                model_card=model_card.replace("{MODEL_DESCRIPTION}",self.opt_model_description)
            else:
                model_card=model_card.replace("{MODEL_DESCRIPTION}","")
                
            ##ADd optional model language
            if self.opt_model_lang is not None:
                if isinstance(self.opt_model_lang,list):
                    model_card = model_card.replace("{LANGUAGE}", "\n".join(["- "+t for t in self.opt_model_lang]))
                else:
                    model_card = model_card.replace("{LANGUAGE}", "\n".join(["- "+t for t in [self.opt_model_lang]]))
            else:
                model_card = model_card.replace("language: \n{LANGUAGE}", "")
                model_card = model_card.replace("It is pretrained for the language : {LANGUAGE}.", "")

            # Add dim info
            self._model_card_vars["{NUM_DIMENSIONS}"] = self.get_sentence_embedding_dimension()

            # Replace vars we created while using the model
            for name, value in self._model_card_vars.items():
                model_card = model_card.replace(name, str(value))

            # Replace remaining vars with default values
            for name, value in ModelCardTemplate.__DEFAULT_VARS__.items():
                model_card = model_card.replace(name, str(value))

        if model_name is not None:
            model_card = model_card.replace("{MODEL_NAME}", model_name.strip())

        with open(os.path.join(path, "README.md"), "w", encoding='utf8') as fOut:
            fOut.write(model_card.strip())

    def save_to_hub(self,
                    repo_name: str,
                    organization: Optional[str] = None,
                    private: Optional[bool] = None,
                    commit_message: str = "Add new LinkTransformer model.",
                    local_model_path: Optional[str] = None,
                    exist_ok: bool = False,
                    replace_model_card: bool = False,
                    train_datasets: Optional[List[str]] = None,
                    override_model_description: Optional[str] = None, 
                    override_model_lang: Optional[str] = None):
        """
        Uploads all elements of this LinkTransformer (inherited Sentence Transformer) to a new HuggingFace Hub repository.

        :param repo_name: Repository name for your model in the Hub.
        :param organization:  Organization in which you want to push your model or tokenizer (you must be a member of this organization).
        :param private: Set to true, for hosting a prive model
        :param commit_message: Message to commit while pushing.
        :param local_model_path: Path of the model locally. If set, this file path will be uploaded. Otherwise, the current model will be uploaded
        :param exist_ok: If true, saving to an existing repository is OK. If false, saving only to a new repository is possible
        :param replace_model_card: If true, replace an existing model card in the hub with the automatically created model card
        :param train_datasets: Datasets used to train the model. If set, the datasets will be added to the model card in the Hub.
        :return: The url of the commit of your model in the given repository.
        """
        token = HfFolder.get_token()
        if token is None:
            raise ValueError("You must login to the Hugging Face hub on this computer by typing `transformers-cli login`.")

        if '/' in repo_name:
            splits = repo_name.split('/', maxsplit=1)
            if organization is None or organization == splits[0]:
                organization = splits[0]
                repo_name = splits[1]
            else:
                raise ValueError("You passed and invalid repository name: {}.".format(repo_name))

        endpoint = "https://huggingface.co"
        repo_id = repo_name
        if organization:
          repo_id = f"{organization}/{repo_id}"
        repo_url = HfApi(endpoint=endpoint).create_repo(
                repo_id=repo_id,
                token=token,
                private=private,
                repo_type=None,
                exist_ok=exist_ok,
            )
        full_model_name = repo_url[len(endpoint)+1:].strip("/")

        with tempfile.TemporaryDirectory() as tmp_dir:
            # First create the repo (and clone its content if it's nonempty).
            logger.info("Create repository and clone it if it exists")
            repo = Repository(tmp_dir, clone_from=repo_url)

            # If user provides local files, copy them.
            if local_model_path:
                copy_tree(local_model_path, tmp_dir)
            else:  # Else, save model directly into local repo.
                create_model_card = replace_model_card or not os.path.exists(os.path.join(tmp_dir, 'README.md'))
                self.save(tmp_dir, model_name=full_model_name, create_model_card=create_model_card, train_datasets=train_datasets)

            ##Save the model config (_lt_model_config) to the repo if it exists
            if hasattr(self, '_lt_model_config'):
                with open(os.path.join(tmp_dir, 'LT_training_config.json'), 'w') as fOut:
                    json.dump(self._lt_model_config, fOut, indent=2)

            #Find files larger 5M and track with git-lfs
            large_files = []
            for root, dirs, files in os.walk(tmp_dir):
                for filename in files:
                    file_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(file_path, tmp_dir)

                    if os.path.getsize(file_path) > (5 * 1024 * 1024):
                        large_files.append(rel_path)

            if len(large_files) > 0:
                logger.info("Track files with git lfs: {}".format(", ".join(large_files)))
                repo.lfs_track(large_files)

            logger.info("Push model to the hub. This might take a while")
            push_return = repo.push_to_hub(commit_message=commit_message)

            def on_rm_error(func, path, exc_info):
                # path contains the path of the file that couldn't be removed
                # let's just assume that it's read-only and unlink it.
                try:
                    os.chmod(path, stat.S_IWRITE)
                    os.unlink(path)
                except:
                    pass

            # Remove .git folder. On Windows, the .git folder might be read-only and cannot be deleted
            # Hence, try to set write permissions on error
            try:
                for f in os.listdir(tmp_dir):
                    shutil.rmtree(os.path.join(tmp_dir, f), onerror=on_rm_error)
            except Exception as e:
                logger.warning("Error when deleting temp folder: {}".format(str(e)))
                pass


        return push_return




class ModelCardTemplate:
    __TAGS__ = ["linktransformer","sentence-transformers", "sentence-similarity","tabular-classification"]
    __DEFAULT_VARS__ = {
        "{PIPELINE_TAG}": "sentence-similarity",
        "{MODEL_DESCRIPTION}": "<!--- Describe your model here -->",
        "{TRAINING_SECTION}": "",
        "{USAGE_TRANSFORMERS_SECTION}": "",
        "{EVALUATION}": "<!--- Describe how your model was evaluated -->",
        "{CITING}": "<!--- Describe where people can find more information -->"
    }

    __MODEL_CARD__ = """
---
pipeline_tag: {PIPELINE_TAG}
language: 
{LANGUAGE}
tags:
{TAGS}
{DATASETS}
---

# {MODEL_NAME}

This is a [LinkTransformer](https://linktransformer.github.io/) model. At its core this model this is a sentence transformer model [sentence-transformers](https://www.SBERT.net) model- it just wraps around the class. 
It is designed for quick and easy record linkage (entity-matching) through the LinkTransformer package. The tasks include clustering, deduplication, linking, aggregation and more.
Notwithstanding that, it can be used for any sentence similarity task within the sentence-transformers framework as well. 
It maps sentences & paragraphs to a {NUM_DIMENSIONS} dimensional dense vector space and can be used for tasks like clustering or semantic search.
Take a look at the documentation of [sentence-transformers](https://www.sbert.net/index.html) if you want to use this model for more than what we support in our applications. 


This model has been fine-tuned on the model : {BASE_MODEL}. It is pretrained for the language : {LANGUAGE}.


{MODEL_DESCRIPTION}

## Usage (LinkTransformer)

Using this model becomes easy when you have [LinkTransformer](https://github.com/dell-research-harvard/linktransformer) installed:

```
pip install -U linktransformer
```

Then you can use the model like this:

```python
import linktransformer as lt
import pandas as pd

##Load the two dataframes that you want to link. For example, 2 dataframes with company names that are written differently
df1=pd.read_csv("data/df1.csv") ###This is the left dataframe with key CompanyName for instance
df2=pd.read_csv("data/df2.csv") ###This is the right dataframe with key CompanyName for instance

###Merge the two dataframes on the key column!
df_merged = lt.merge(df1, df2, on="CompanyName", how="inner")

##Done! The merged dataframe has a column called "score" that contains the similarity score between the two company names

```


## Training your own LinkTransformer model
Any Sentence Transformers can be used as a backbone by simply adding a pooling layer.  Any other transformer on HuggingFace can also be used by specifying the option add_pooling_layer==True
The model was trained using SupCon loss. 
Usage can be found in the package docs. 
The training config can be found in the repo with the name LT_training_config.json
To replicate the training, you can download the file and specify the path in the config_path argument of the training function. You can also override the config by specifying the training_args argument.
Here is an example. 


```python

##Consider the example in the paper that has a dataset of Mexican products and their tariff codes from 1947 and 1948 and we want train a model to link the two tariff codes.
saved_model_path = train_model(
        model_path="hiiamsid/sentence_similarity_spanish_es",
        dataset_path=dataset_path,
        left_col_names=["description47"],
        right_col_names=['description48'],
        left_id_name=['tariffcode47'],
        right_id_name=['tariffcode48'],
        log_wandb=False,
        config_path=LINKAGE_CONFIG_PATH,
        training_args={"num_epochs": 1}
    )

```


You can also use this package for deduplication (clusters a df on the supplied key column). Merging a fine class (like product) to a coarse class (like HS code) is also possible.
Read our paper and the documentation for more!

{USAGE_TRANSFORMERS_SECTION}

## Evaluation Results

{EVALUATION}

You can evaluate the model using the [LinkTransformer](https://github.com/dell-research-harvard/linktransformer) package's inference functions.
We have provided a few datasets in the package for you to try out. We plan to host more datasets on Huggingface and our website (Coming soon) that you can take a look at.

{TRAINING_SECTION}



{FULL_MODEL_STR}
```

## Citing & Authors

```
@misc{arora2023linktransformer,
                  title={LinkTransformer: A Unified Package for Record Linkage with Transformer Language Models},
                  author={Abhishek Arora and Melissa Dell},
                  year={2023},
                  eprint={2309.00789},
                  archivePrefix={arXiv},
                  primaryClass={cs.CL}
                }

```

"""



    __TRAINING_SECTION__ = """
## Training

### Training your own LinkTransformer model
Any Sentence Transformers can be used as a backbone by simply adding a pooling layer.  Any other transformer on HuggingFace can also be used by specifying the option add_pooling_layer==True
The model was trained using SupCon loss. 
Usage can be found in the package docs. 
The training config can be found in the repo with the name LT_training_config.json
To replicate the training, you can download the file and specify the path in the config_path argument of the training function. You can also override the config by specifying the training_args argument.
Here is an example. 


```python

##Consider the example in the paper that has a dataset of Mexican products and their tariff codes from 1947 and 1948 and we want train a model to link the two tariff codes.
saved_model_path = train_model(
        model_path="hiiamsid/sentence_similarity_spanish_es",
        dataset_path=dataset_path,
        left_col_names=["description47"],
        right_col_names=['description48'],
        left_id_name=['tariffcode47'],
        right_id_name=['tariffcode48'],
        log_wandb=False,
        config_path=LINKAGE_CONFIG_PATH,
        training_args={"num_epochs": 1}
    )


```



"""


class ClassificationModelCardTemplate:
    __TAGS__ = ["linktransformer", "transformers", "text-classification", "tabular-classification"]
    __DEFAULT_VARS__ = {
        "{PIPELINE_TAG}": "text-classification",
        "{MODEL_DESCRIPTION}": "<!--- Describe your model here -->",
        "{TRAINING_SECTION}": "",
        "{USAGE_TRANSFORMERS_SECTION}": "",
        "{EVALUATION}": "<!--- Describe how your model was evaluated -->",
        "{CITING}": "<!--- Describe where people can find more information -->"
    }

    __MODEL_CARD__ = """
---
pipeline_tag: {PIPELINE_TAG}
language: 
{LANGUAGE}
tags:
{TAGS}
{DATASETS}
---

# {MODEL_NAME}

This model is part of the [LinkTransformer](https://linktransformer.github.io/) ecosystem. While rooted in the a standard HuggingFace Transformer, this specific instance is tailored for text classification tasks. It classifies input sentences or paragraphs into specific categories or labels, leveraging the power of transformer architectures.

The base model for this classifier is: {BASE_MODEL}. It is pretrained for the language: {LANGUAGE}.

Labels are mapped to integers as follows: 

{LABEL_MAP}

{MODEL_DESCRIPTION}

## Usage with LinkTransformer

After installing [LinkTransformer](https://linktransformer.github.io/):

```python
pip install -U linktransformer
```

Employ the model for text classification tasks:

```python
import linktransformer as lt
df_clf_output = lt.classify_rows(df, on=["col_of_interest"], model="{MODEL_NAME}")
```


## Training

### Training your own LinkTransformer Classification Model

With the provided tools, you can train a custom classification model:

```python
from linktransformer import train_clf_model

best_model_path, best_metric, label_map = train_clf_model(
    data="path_to_dataset.csv",
    model="you-model-path-or-name",
    on=["col_of_interest"],
    label_col_name="label_column_name",
    lr=5e-5,
    batch_size=16,
    epochs=3
)
```

{TRAINING_SECTION}

{USAGE_TRANSFORMERS_SECTION}

## Evaluation Results

{EVALUATION}

Evaluation is typically based on metrics like accuracy, F1-score, precision, and recall.

## Citing & Authors

```
@misc{arora2023linktransformer,
                  title={LinkTransformer: A Unified Package for Record Linkage with Transformer Language Models},
                  author={Abhishek Arora and Melissa Dell},
                  year={2023},
                  eprint={2309.00789},
                  archivePrefix={arXiv},
                  primaryClass={cs.CL}
}
```

"""

    __TRAINING_SECTION__ = """
## Training

### Training your own LinkTransformer Classification Model

With the provided tools, you can train a custom classification model:

```python
from linktransformer import train_clf_model

best_model_path, best_metric, label_map = train_clf_model(
    data="path_to_dataset.csv",
    on=["text_column_name"],
    label_col_name="label_column_name",
    lr=5e-5,
    batch_size=16,
    epochs=3
)
```

More information on configurations can be found in the [LinkTransformer documentation](https://linktransformer.github.io/).
"""




import math
import torch
import torch.nn as nn
from sentence_transformers import LoggingHandler
import logging
from transformers import logging as lg
import wandb


lg.set_verbosity_error()
logging.basicConfig(format='%(asctime)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    level=logging.INFO,
                    handlers=[LoggingHandler()])
logger = logging.getLogger(__name__)






class SupConLoss_wandb(nn.Module):  
    """
    Source: https://github.com/UKPLab/sentence-transformers/issues/1604
    Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    """

    def __init__(self, model, temperature=0.07, contrast_mode='all',
                 base_temperature=0.07, subbatch_size=16,wandb_names=None):
        super(SupConLoss_wandb, self).__init__()
        self.model = model
        self.temperature = temperature
        self.contrast_mode = contrast_mode
        self.base_temperature = base_temperature
        self.subbatch_size = subbatch_size
        self.wandb_names = wandb_names

    def forward(self, sentence_features, labels=None, mask=None):
        """Compute loss for model. If both `labels` and `mask` are None,
        it degenerates to SimCLR unsupervised loss:
        https://arxiv.org/pdf/2002.05709.pdf
        Args:
            features: hidden vector of shape [batch size, n_views, ...].
            labels: ground truth of shape [batch size].
            mask: contrastive mask of shape [batch size, batch size], mask_{i,j}=1 if sample j
                has the same class as sample i. Can be asymmetric.
        Returns:
            A loss scalar.
        """
        features = self.model(sentence_features[0])['sentence_embedding']

        #Nils: Normalize embeddings
        features = torch.nn.functional.normalize(features, p=2, dim=1)

        ## Nils: Add n_views dimension
        features = torch.unsqueeze(features, 1)

        device = features.device

        if len(features.shape) < 3:
            raise ValueError('`features` needs to be [bsz, n_views, ...],'
                             'at least 3 dimensions are required')
        if len(features.shape) > 3:
            features = features.view(features.shape[0], features.shape[1], -1)

        batch_size = features.shape[0]
        if labels is not None and mask is not None:
            raise ValueError('Cannot define both `labels` and `mask`')
        elif labels is None and mask is None:
            mask = torch.eye(batch_size, dtype=torch.float32).to(device)
        elif labels is not None:
            labels = labels.contiguous().view(-1, 1)
            if labels.shape[0] != batch_size:
                raise ValueError('Num of labels does not match num of features')
            mask = torch.eq(labels, labels.T).float().to(device)
        else:
            mask = mask.float().to(device)

        contrast_count = features.shape[1]
        contrast_feature = torch.cat(torch.unbind(features, dim=1), dim=0)
        if self.contrast_mode == 'one':
            anchor_feature = features[:, 0]
            anchor_count = 1
        elif self.contrast_mode == 'all':
            anchor_feature = contrast_feature
            anchor_count = contrast_count
        else:
            raise ValueError('Unknown mode: {}'.format(self.contrast_mode))


        # compute logits
        anchor_dot_contrast = torch.div(                           # dot prods bewteen every pair of vectors
            torch.matmul(anchor_feature, contrast_feature.T),
            self.temperature)

        # for numerical stability
        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        # tile mask
        mask = mask.repeat(anchor_count, contrast_count)
        # mask-out self-contrast cases
        logits_mask = torch.scatter(
            torch.ones_like(mask),
            1,
            torch.arange(batch_size * anchor_count).view(-1, 1).to(device),
            0
        )
        mask = mask * logits_mask

        # compute log_prob
        exp_logits = torch.exp(logits) * logits_mask
        log_prob = logits - torch.log(exp_logits.sum(1, keepdim=True))
        
        # compute mean of log-likelihood over positive
        mean_log_prob_pos = (mask * log_prob).sum(1) / mask.sum(1)

        # loss
        loss = - (self.temperature / self.base_temperature) * mean_log_prob_pos
        loss = loss.view(anchor_count, batch_size).mean()
        if self.wandb_names is not None:
            wandb.log({"Loss": loss}) 

        if math.isnan(loss):
            raise ValueError("Loss")

        return loss



###Inference and Linkage script
###We want to link dfs together using embeddings
import json
import os
import warnings

import numpy as np
import pandas as pd
import faiss
from typing import Union, List, Optional, Tuple,Dict, Any
from pandas import DataFrame

from linktransformer.modified_sbert.cluster_fns import cluster
# from linktransformer.utils import serialize_columns, infer_embeddings, load_model, load_clf, cosine_similarity_corresponding_pairs, tokenize_data_for_inference, predict_rows_with_openai
from linktransformer.utils import *
from sklearn.metrics.pairwise import cosine_similarity
from itertools import combinations
from transformers import TrainingArguments, Trainer



def merge(
    df1: DataFrame,
    df2: DataFrame,
    merge_type: str = '1:1',
    on: Optional[Union[str, List[str]]] = None,
    model: str = "all-MiniLM-L6-v2",
    left_on: Optional[Union[str, List[str]]] = None,
    right_on: Optional[Union[str, List[str]]] = None,
    suffixes: Tuple[str, str] = ('_x', '_y'),
    use_gpu: bool = False,
    batch_size: int = 128,
    openai_key: Optional[str] = None
) -> DataFrame:
    """
    Merge two dataframes using language model embeddings.

    :param df1 (DataFrame): First dataframe (left).
    :param df2 (DataFrame): Second dataframe (right).
    :param merge_type (str): Type of merge to perform (1:m or m:1 or 1:1).
    :param model (str): Language model to use.
    :param on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param left_on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param right_on (Union[str, List[str]], optional): Column(s) to join on in df2. Defaults to None.
    :param suffixes (Tuple[str, str]): Suffixes to use for overlapping columns. Defaults to ('_x', '_y').
    :param use_gpu (bool): Whether to use GPU. Not supported yet. Defaults to False.
    :param batch_size (int): Batch size for inferencing embeddings. Defaults to 128.
    :param openai_key (str, optional): OpenAI API key for InferKit API. Defaults to None.
    :return: DataFrame: The merged dataframe.
    """
    ## Set common columns as on if not specified
    if on is None:
        on = list(set(df1.columns).intersection(set(df2.columns)))

    ## If left_on or right_on is not specified, set it to on
    if left_on is None:
        left_on = on
    if right_on is None:
        right_on = on

    on = None

    ### If how is 1:m, then we want to merge df1 to df2
    ### Check if how is valid
    if merge_type not in ["1:m", "m:1", "1:1"]:
        raise ValueError(f"Invalid merge type: {merge_type}")

    if merge_type == "1:m":
        if df1[left_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")

    if merge_type == "m:1":
        ## Check if keys in df2 are unique
        if df2[right_on].duplicated().any():
            print("Warning: Keys in df2 are not unique")
    if merge_type == "1:1":
        ## Check if keys in df1 are unique
        if df1[left_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")
        ## Check if keys in df2 are unique
        if df2[right_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")

    df1 = df1.copy()
    df2 = df2.copy()
    ## give ids to each df
    ##Ensure that there is no id_lt column in df1 or df2
    if "id_lt" in df1.columns:
        raise ValueError(f"Column id_lt already exists in df1, please rename it to proceed")
    if "id_lt" in df2.columns:
        raise ValueError(f"Column id_lt already exists in df2,please rename it to proceed")

    df1.loc[:, "id_lt"] = np.arange(len(df1))
    df2.loc[:, "id_lt"] = np.arange(len(df2))

    if isinstance(right_on, list):
        strings_right = serialize_columns(df2, right_on, model=model)
    if isinstance(left_on, list):
        strings_left = serialize_columns(df1, left_on, model=model)
    else:
        strings_left = df1[left_on].tolist()
        strings_right = df2[right_on].tolist()

    ## Load the model
    if openai_key is None:
        model = load_model(model)

    ## Infer embeddings for df1
    embeddings1 = infer_embeddings(strings_left, model, batch_size=batch_size, openai_key=openai_key)
    ## Infer embeddings for df2
    embeddings2 = infer_embeddings(strings_right, model, batch_size=batch_size, openai_key=openai_key)

    ### Expand dim if embeddings are 1d (numpy)
    if len(embeddings1.shape) == 1:
        embeddings1 = np.expand_dims(embeddings1, axis=0)
    if len(embeddings2.shape) == 1:
        embeddings2 = np.expand_dims(embeddings2, axis=0)
    ## Normalize embedding tensors using numpy

    embeddings1 = embeddings1 / np.linalg.norm(embeddings1, axis=1, keepdims=True)
    embeddings2 = embeddings2 / np.linalg.norm(embeddings2, axis=1, keepdims=True)
    ## Create index
    if use_gpu:
        raise ValueError(f"GPU not supported yet")
    else:
        index = faiss.IndexFlatIP(embeddings1.shape[1])

    ## Add to index depending on merge type
    if merge_type == "1:m":
        index.add(embeddings2)
    elif merge_type == "m:1":
        index.add(embeddings2)
    elif merge_type == "1:1":
        index.add(embeddings2)

    ## Search index
    if merge_type == "1:m":
        D, I = index.search(embeddings1, 1)
    elif merge_type == "m:1":
        D, I = index.search(embeddings1, 1)
    elif merge_type == "1:1":
        D, I = index.search(embeddings1, 1)

    ## Check nearest neighbor of the first text in df1 as a test
    df1 = df1.reset_index(drop=True)
    df2 = df2.reset_index(drop=True)

    ## Fuzzily merge the dfs based on the faiss index queries
    df_lm_matched = df1.merge(df2.iloc[I.flatten()].reset_index(drop=True), left_index=True, right_index=True, how="inner")
    ### Add score column
    df_lm_matched["score"] = D.flatten()

    print(f"LM matched on key columns - left: {left_on}{suffixes[0]}, right: {right_on}{suffixes[1]}")
    return df_lm_matched

    


def merge_blocking(
    df1: DataFrame,
    df2: DataFrame,
    merge_type: str = '1:1',
    on: Optional[Union[str, List[str]]] = None,
    model: str = "all-MiniLM-L6-v2",
    left_on: Optional[Union[str, List[str]]] = None,
    right_on: Optional[Union[str, List[str]]] = None,
    blocking_vars: Optional[List[str]] = None,
    suffixes: Tuple[str, str] = ('_x', '_y'),
    use_gpu: bool = False,
    batch_size: int = 128,
    openai_key: Optional[str] = None
) -> DataFrame:
    """
    Merge two dataframes using language model embeddings with optional blocking.

    :param df1 (DataFrame): First dataframe (left).
    :param df2 (DataFrame): Second dataframe (right).
    :param merge_type (str): Type of merge to perform (1:m or m:1 or 1:1).
    :param model (str): Language model to use.
    :param on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param left_on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param right_on (Union[str, List[str]], optional): Column(s) to join on in df2. Defaults to None.
    :param blocking_vars (List[str], optional): Columns to use for blocking. Defaults to None.
    :param suffixes (Tuple[str, str]): Suffixes to use for overlapping columns. Defaults to ('_x', '_y').
    :param use_gpu (bool): Whether to use GPU. Not supported yet. Defaults to False.
    :param batch_size (int): Batch size for inferencing embeddings. Defaults to 128.
    :param openai_key (str, optional): OpenAI API key for InferKit API. Defaults to None.
    :return: DataFrame: The merged dataframe.
    """
    ### For blocking, we need to chunk the dfs into blocks
    ### First, we need to check if blocking vars are specified
    if blocking_vars is None:
        print("No blocking vars specified, matching between all rows")
        df_lm_matched = merge(df1, df2, merge_type=merge_type, on=on, model=model, left_on=left_on,
                                    right_on=right_on, suffixes=suffixes, use_gpu=use_gpu, batch_size=batch_size,
                                     openai_key=openai_key)
        return df_lm_matched
    else:
        ## Partition the dfs into blocks
        ## First, check if blocking vars are in df1 and df2 - both should exist
        if not set(blocking_vars).issubset(set(df1.columns)):
            raise ValueError(f"Blocking vars {blocking_vars} not in df1")
        if not set(blocking_vars).issubset(set(df2.columns)):
            raise ValueError(f"Blocking vars {blocking_vars} not in df2")

        ## Now, chunk the dfs into blocks
        ## First, get the number of blocks
        num_blocks_1 = len(df1.groupby(blocking_vars))
        num_blocks_2 = len(df2.groupby(blocking_vars))
        print(f"Number of blocks in df1: {num_blocks_1}")
        print(f"Number of blocks in df2: {num_blocks_2}")

        ## Now, get the blocks
        df1_blocks = df1.groupby(blocking_vars)
        df2_blocks = df2.groupby(blocking_vars)

        ## Now, we need to merge each block in df1 with each block in df2.
        ## We need to keep track of the blocks that have been merged

        ## First, create a list of all the blocks in df1 and df2
        df1_block_list = list(df1_blocks.groups.keys())
        df2_block_list = list(df2_blocks.groups.keys())

        ### Iterate through each block in df1. Check if the block is in df2. If it is, merge the blocks.
        ### If not, add keys to skipped keys list
        merged_dfs = []
        skipped_keys_1 = set(df1_block_list).difference(set(df2_block_list))
        skipped_keys_2 = set(df2_block_list).difference(set(df1_block_list))
        common_keys = set(df1_block_list).intersection(set(df2_block_list))

        for block_1 in common_keys:
            print(f"Merging block {block_1}")
            df1_block = df1_blocks.get_group(block_1)
            df2_block = df2_blocks.get_group(block_1)
            ## Merge the blocks
            df_block_matched = merge(df1_block, df2_block, merge_type=merge_type, on=on, model=model,
                                           left_on=left_on, right_on=right_on, suffixes=suffixes, use_gpu=use_gpu,
                                           batch_size=batch_size, openai_key=openai_key)
            ## Add to merged dfs
            merged_dfs.append(df_block_matched)

        ## Add df corresponding to skipped keys
        for block_1 in skipped_keys_1:
            df1_block = df1_blocks.get_group(block_1)
            merged_dfs.append(df1_block)

        for block_2 in skipped_keys_2:
            df2_block = df2_blocks.get_group(block_2)
            merged_dfs.append(df2_block)

        ## Concatenate the merged dfs
        df_lm_matched = pd.concat(merged_dfs, axis=0).reset_index(drop=True)
        return df_lm_matched




def aggregate_rows(
    df: DataFrame,
    ref_df: DataFrame,
    model: str,
    left_on: Union[str, List[str]],
    right_on: Union[str, List[str]],
    openai_key: str = None
) -> DataFrame:
    """
    Aggregate the dataframe based on a reference dataframe using a language model.

    :param df (DataFrame): Dataframe to aggregate.
    :param ref_df (DataFrame): Reference dataframe to aggregate on.
    :param model (str): Language model to use.
    :param left_on (Union[str, List[str]]): Column(s) to aggregate on in df.
    :param right_on (Union[str, List[str]]): Reference column(s) to aggregate on in ref_df.
    :return: DataFrame: The aggregated dataframe.
    """

    df = df.copy()
    ref_df = ref_df.copy()

    ## Just use the merge function with merge type 1:m
    df_lm_matched = merge(df, ref_df, merge_type="1:1", on=None, model=model, left_on=left_on,
                                right_on=right_on, suffixes=("_x", "_y"), use_gpu=False, batch_size=128,
                                 openai_key=openai_key)

    return df_lm_matched



def evaluate_pairs(df,model,left_on,right_on,openai_key=None):
    """
    This function evaluates paired columns in a dataframe and gives a match score (cosine similarity). 
    Typically, this can be though of as a way to evaluate already merged in dataframes.

    :param df (DataFrame): Dataframe to evaluate.
    :param model (str): Language model to use.
    :param left_on (Union[str, List[str]]): Column(s) to evaluate on in df.
    :param right_on (Union[str, List[str]]): Reference column(s) to evaluate on in df.
    :return: DataFrame: The evaluated dataframe.
    """

    df = df.copy()

    ###We will serialize the columns if they are lists
    if isinstance(left_on, list):
        strings_left = serialize_columns(df, left_on, model=model)
    else:
        strings_left = df[left_on].tolist()
    
    if isinstance(right_on, list):
        strings_right = serialize_columns(df, right_on, model=model)
    else:
        strings_right = df[right_on].tolist()

    ## Load the model
    if openai_key is None:
        model = load_model(model)

    ## Infer embeddings for df1
    embeddings1 = infer_embeddings(strings_left, model, batch_size=128, openai_key=openai_key)
    ## Infer embeddings for df2
    embeddings2 = infer_embeddings(strings_right, model, batch_size=128, openai_key=openai_key)

    ### Expand dim if embeddings are 1d (numpy)
    if len(embeddings1.shape) == 1:
        embeddings1 = np.expand_dims(embeddings1, axis=0)
    if len(embeddings2.shape) == 1:
        embeddings2 = np.expand_dims(embeddings2, axis=0)
    ## Normalize embedding tensors using numpy

    embeddings1 = embeddings1 / np.linalg.norm(embeddings1, axis=1, keepdims=True)
    embeddings2 = embeddings2 / np.linalg.norm(embeddings2, axis=1, keepdims=True)

    ## Compute cosine similarity between CORRESPONDING pair of embeddings
    cosine_similarity_12 = cosine_similarity_corresponding_pairs(embeddings1,embeddings2)

    ## Add cosine similarity to df
    df["score"] = cosine_similarity_12.flatten()

    return df

def cluster_rows(
    df: DataFrame,
    model: str,
    on: Union[str, List[str]],
    cluster_type: str = "SLINK",
    cluster_params: Dict[str, Any] = {'threshold': 0.5, "min cluster size": 2, "metric": "cosine"},
    openai_key: str = None
) -> DataFrame:
    """
    Deduplicate a dataframe based on a similarity threshold. Various clustering options are supported.         
    "agglomerative": {
            "threshold": 0.5,
            "clustering linkage": "ward",  # You can choose a default linkage method
            "metric": "euclidean",  # You can choose a default metric
        },
        "HDBScan": {
            "min cluster size": 5,
            "min samples": 1,
        },
        "SLINK": {
            "min cluster size": 2,
            "threshold": 0.1,
        },
    }

    :param df (DataFrame): Dataframe to deduplicate.
    :param model (str): Language model to use.
    :param on (Union[str, List[str]]): Column(s) to deduplicate on.
    :param cluster_type (str): Clustering method to use. Defaults to "SLINK".
    :param cluster_params (Dict[str, Any]): Parameters for clustering method. Defaults to {'threshold': 0.5, "min cluster size": 2, "metric": "cosine"}.
    :param openai_key (str): OpenAI API key
    :return: DataFrame: The deduplicated dataframe.
    """

    df = df.copy()


    ## First, get the embeddings
    ### If len(on)>1, then we need to serialize the columns
    if isinstance(on, list):
        strings = serialize_columns(df, on, model=model)
    else:
        strings = df[on].tolist()
    
    ## Infer embeddings for df
    embeddings = infer_embeddings(strings, model, batch_size=128, openai_key=openai_key)
    ## Normalize embedding tensors using numpy
    embeddings = embeddings / np.linalg.norm(embeddings, axis=1, keepdims=True)
    ### Now, cluster the embeddings based on similarity threshold
    labels = cluster(cluster_type, cluster_params, embeddings, corpus_ids=None)
    ### Now, keep only 1 row per cluster
    df["cluster"] = labels
    return df





def dedup_rows(
    df: DataFrame,
    model: str,
    on: Union[str, List[str]],
    cluster_type: str = "SLINK",
    cluster_params: Dict[str, Any] = {'threshold': 0.5, "min cluster size": 2, "metric": "cosine"},
    openai_key: str = None
) -> DataFrame:
    """
    Deduplicate a dataframe based on a similarity threshold. This is just clustering and keeping the first row in each cluster.
    Refer to the docs for the cluster_rows function for more details.

    :param df (DataFrame): Dataframe to deduplicate.
    :param model (str): Language model to use.
    :param on (Union[str, List[str]]): Column(s) to deduplicate on.
    :param cluster_type (str): Clustering method to use. Defaults to "SLINK".
    :param cluster_params (Dict[str, Any]): Parameters for clustering method. Defaults to {'threshold': 0.5, "min cluster size": 2, "metric": "cosine"}.
    :param openai_key (str): OpenAI API key
    :return: DataFrame: The deduplicated dataframe.
    """

    df = df.copy()

    

    print(f"Deduplicating dataframe with originally {len(df)} rows")
    ##Drop exact duplicates
    print("Checking for and dropping exact duplicates")
    df = df.drop_duplicates(subset=on, keep="first")
    print(f"Number of rows after dropping exact duplicates: {len(df)}")

    df = cluster_rows(df, model, on, cluster_type, cluster_params, openai_key)
    df = df.drop_duplicates(subset="cluster", keep="first")
    df = df.drop(columns=["cluster"])
    print(f"Number of rows after deduplication: {len(df)}")

    return df


    

def all_pair_combos_evaluate(df,model,left_on,right_on,openai_key=None):
    """
    Get similarity scores for every pair of rows in a dataframe. 
    We make this efficient by only embedding each string once and get all possible pairwise distances
    and add the expanded rows and their scores to the dataframe
    :param df (DataFrame): Dataframe to evaluate.
    :param model (str): Language model to use.
    :param left_on (Union[str, List[str]]): Column(s) to evaluate on in df.
    :param right_on (Union[str, List[str]]): Reference column(s) to evaluate on in df.
    :param openai_key (str): OpenAI API key
    :return: DataFrame: The evaluated dataframe.
    """

    ###Get the embeddings for the left_on column
    if isinstance(left_on, list):
        strings_left = serialize_columns(df, left_on, model=model)
    else:
        strings_left = df[left_on].tolist()
    
    ## Infer embeddings for df1
    embeddings1 = infer_embeddings(strings_left, model, batch_size=128, openai_key=openai_key)
    embeddings1 = embeddings1 / np.linalg.norm(embeddings1, axis=1, keepdims=True)

    ###Get the embeddings for the right_on column
    if isinstance(right_on, list):
        strings_right = serialize_columns(df, right_on, model=model)
    else:
        strings_right = df[right_on].tolist()
    
    ## Infer embeddings for df1
    embeddings2 = infer_embeddings(strings_right, model, batch_size=128, openai_key=openai_key)
    embeddings2 = embeddings2 / np.linalg.norm(embeddings2, axis=1, keepdims=True)

    ###calculate the cosine similarity between all pairs of rows
    cosine_similarity_12_all_pairs = cosine_similarity(embeddings1,embeddings2)
    ##Flatten the matrix
    cosine_similarity_12_all_pairs = cosine_similarity_12_all_pairs.flatten()

    ###This gives an n*n matrix of cosine similarities. 

    ###Similarly, make an n*n matrix of the left_on column and right_on column
    left_on_all_pairs = np.repeat(df[left_on].values, len(df[right_on]), axis=0)


    ###Left part of the df was repeating n times. We also want the right one to repeat, but 


    right_on_all_pairs = np.tile(df[right_on].values, (len(df[left_on]),1))

    ###Now, flattenn the right on pairs
    right_on_all_pairs = right_on_all_pairs.flatten()

    ###Now, we can make a dataframe with the left_on, right_on and cosine similarity
    df_concat = pd.DataFrame({"left_on":left_on_all_pairs,"right_on":right_on_all_pairs,"score":cosine_similarity_12_all_pairs})

    return df_concat



def merge_knn(
    df1: DataFrame,
    df2: DataFrame,
    merge_type: str = '1:1',
    on: Optional[Union[str, List[str]]] = None,
    model: str = "all-MiniLM-L6-v2",
    left_on: Optional[Union[str, List[str]]] = None,
    right_on: Optional[Union[str, List[str]]] = None,
    k: int = 1,
    suffixes: Tuple[str, str] = ('_x', '_y'),
    use_gpu: bool = False,
    batch_size: int = 128,
    openai_key: Optional[str] = None,
    drop_sim_threshold: float = None
) -> DataFrame:
    """
    Merge two dataframes using language model embeddings. This function would support k nearest neighbors matching for each row in df1.
    Merge is a special case of this function when k=1.
    :param df1 (DataFrame): First dataframe (left).
    :param df2 (DataFrame): Second dataframe (right).
    :param on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param model (str): Language model to use.
    :param left_on (Union[str, List[str]], optional): Column(s) to join on in df1. Defaults to None.
    :param right_on (Union[str, List[str]], optional): Column(s) to join on in df2. Defaults to None.
    :param k (int): Number of nearest neighbors to match for each row in df1. Defaults to 1.
    :param suffixes (Tuple[str, str]): Suffixes to use for overlapping columns. Defaults to ('_x', '_y').
    :param use_gpu (bool): Whether to use GPU. Not supported yet. Defaults to False.
    :param batch_size (int): Batch size for inferencing embeddings. Defaults to 128.
    :param openai_key (str, optional): OpenAI API key for InferKit API. Defaults to None.
    :return: DataFrame: The merged dataframe.
    """

    ## Set common columns as on if not specified
    if on is None:
        on = list(set(df1.columns).intersection(set(df2.columns)))

    ## If left_on or right_on is not specified, set it to on
    if left_on is None:
        left_on = on
    if right_on is None:
        right_on = on

    on = None

    ### If how is 1:m, then we want to merge df1 to df2
    ### Check if how is valid
    if merge_type not in ["1:m", "m:1", "1:1"]:
        raise ValueError(f"Invalid merge type: {merge_type}")

    if merge_type == "1:m":
        if df1[left_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")

    if merge_type == "m:1":
        ## Check if keys in df2 are unique
        if df2[right_on].duplicated().any():
            print("Warning: Keys in df2 are not unique")
    if merge_type == "1:1":
        ## Check if keys in df1 are unique
        if df1[left_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")
        ## Check if keys in df2 are unique
        if df2[right_on].duplicated().any():
           print("Warning: Keys in df1 are not unique")

    df1 = df1.copy()
    df2 = df2.copy()
    ## give ids to each df
    ##Ensure that there is no id_lt column in df1 or df2
    if "id_lt" in df1.columns:
        raise ValueError(f"Column id_lt already exists in df1, please rename it to proceed")
    if "id_lt" in df2.columns:
        raise ValueError(f"Column id_lt already exists in df2,please rename it to proceed")

    df1.loc[:, "id_lt"] = np.arange(len(df1))
    df2.loc[:, "id_lt"] = np.arange(len(df2))

    if isinstance(right_on, list):
        strings_right = serialize_columns(df2, right_on, model=model)
    if isinstance(left_on, list):
        strings_left = serialize_columns(df1, left_on, model=model)
    else:
        strings_left = df1[left_on].tolist()
        strings_right = df2[right_on].tolist()
    
    ## Load the model
    model = load_model(model)

    ## Infer embeddings for df1
    embeddings1 = infer_embeddings(strings_left, model, batch_size=batch_size, openai_key=openai_key)
    ## Infer embeddings for df2
    embeddings2 = infer_embeddings(strings_right, model, batch_size=batch_size, openai_key=openai_key)

    ### Expand dim if embeddings are 1d (numpy)
    if len(embeddings1.shape) == 1:
        embeddings1 = np.expand_dims(embeddings1, axis=0)
    if len(embeddings2.shape) == 1:
        embeddings2 = np.expand_dims(embeddings2, axis=0)

    ## Normalize embedding tensors using numpy

    embeddings1 = embeddings1 / np.linalg.norm(embeddings1, axis=1, keepdims=True)
    embeddings2 = embeddings2 / np.linalg.norm(embeddings2, axis=1, keepdims=True)

    ## Create index
    if use_gpu:
        raise ValueError(f"GPU not supported yet")
    else:
        index = faiss.IndexFlatIP(embeddings1.shape[1])

    ## Add to index depending on merge type
    if merge_type == "1:m":
        index.add(embeddings2)
    elif merge_type == "m:1":
        index.add(embeddings2)
    elif merge_type == "1:1":
        index.add(embeddings2)

    ## Search index
    if merge_type == "1:m":
        D, I = index.search(embeddings1, k)
    elif merge_type == "m:1":
        D, I = index.search(embeddings1, k)
    elif merge_type == "1:1":
        D, I = index.search(embeddings1, k)

    ## Check nearest neighbor of the first text in df1 as a test
    df1 = df1.reset_index(drop=True)
    df2 = df2.reset_index(drop=True)

    ## Fuzzily merge the dfs based on the faiss index queries
    ###Each I sublst is a list of k nearest neighbors for each row in df1 in terms of indices of df2
    ###We need to expand the rows of df1 and df2 to match the number of rows in df1
    ###We also need to expand the scores to match the number of rows in df1

    ###First, expand the rows of df1
    df1_expanded = df1.loc[np.repeat(df1.index.values, k)].reset_index(drop=True)

    ###Now, expand the rows of df2
    df2_expanded = df2.iloc[I.flatten()].reset_index(drop=True)


    ###Now, merge the expanded dfs
    df_lm_matched = df1_expanded.merge(df2_expanded, left_index=True, right_index=True, how="inner")

    ### Add score column
    df_lm_matched["score"] =  D.flatten()

    print(f"LM matched on key columns - left: {left_on}{suffixes[0]}, right: {right_on}{suffixes[1]}")

    if drop_sim_threshold is not None:
        df_lm_matched = df_lm_matched[df_lm_matched["score"]>=drop_sim_threshold]
        print(f"Dropped rows with similarity below {drop_sim_threshold}")

    return df_lm_matched


def classify_rows(
    df: DataFrame,
    on: Optional[Union[str, List[str]]] = None,
    model: str = None,
    num_labels: int = 2,
    label_map: Optional[dict] = None,
    use_gpu: bool = False,
    batch_size: int = 128,
    openai_key: Optional[str] = None,
    openai_topic: Optional[str] = None,
    openai_prompt: Optional[str] = None,
    openai_params: Optional[dict] = {}
):
    """
    Classify texts in all rows of one or more columns whether they are relevant to a certain topic. The function uses
    either a trained classifier to make predictions or an OpenAI API key to send requests and retrieve classification
    results from ChatCompletion endpoint. The function returns a copy of the input dataframe with a new column "clf_preds_{on}" that stores the
    classification results.

    :param df: (DataFrame) the dataframe.
    :param on: (Union[str, List[str]], optional) Column(s) to classify (if multiple columns are passed in, they will be joined).
    :param model: (str) filepath to the model to use (to use OpenAI, see "https://platform.openai.com/docs/models").
    :param num_labels: (int) number of labels to predict. Defaults to 2.
    :param label_map: (dict) a dictionary that maps text labels to numeric labels. Used for OpenAI predictions.
    :param use_gpu: (bool) Whether to use GPU. Not supported yet. Defaults to False.
    :param batch_size: (int) Batch size for inferencing embeddings. Defaults to 128.
    :param openai_key: (str, optional) OpenAI API key for InferKit API. Defaults to None.
    :param openai_topic: (str, optional) The topic predict whether the text is relevant or not. Defaults to None.
    :param openai_prompt: (str, optional) Custom system prompt for OpenAI ChatCompletion endpoint. Defaults to None.
    :param openai_params: (str, optional) Custom parameters for OpenAI ChatCompletion endpoint. Defaults to None.
    :returns: DataFrame: The dataframe with a new column "clf_preds_{on}" that stores the classification results.
    """

    df = df.copy()

    # load label dict from model path if exists
    if os.path.exists(os.path.join(model, "label_map.json")):
        with open(os.path.join(model, "label_map.json"), encoding="utf-8") as f:
            label_map = json.load(f)

    # check if label dict is compatible with num of labels
    if label_map is not None and len(label_map) != num_labels:
        raise ValueError(f"label_dict has {label_map} entries, but num_labels is {num_labels}. ")

    # check if the parameters for any inference method is properly specified
    if openai_key is None and model is None:
        raise ValueError("Must either specify a model or use an OpenAI key")

    # infer using HF or local models
    if openai_key is None:
        clf = load_clf(model, num_labels=num_labels)

        # join texts in columns if needed
        if isinstance(on, list):
            strings_col = serialize_columns(df, on, model=model)
        else:
            strings_col = df[on].tolist()

        # tokenize data
        tokenized_data = tokenize_data_for_inference(strings_col, "inf_data", model)

        # initialize trainer
        inference_args = TrainingArguments(output_dir="save", per_device_eval_batch_size=batch_size)
        trainer = Trainer(model=clf, args=inference_args)

        # predict and save results
        predictions = trainer.predict(tokenized_data)
        preds = np.argmax(predictions.predictions, axis=-1)

        # convert numeric labels to text labels according to label_map if exists
        if label_map is not None:
            reversed_label_map = {val: key for (key, val) in label_map.items()}
            try:
                preds = [reversed_label_map[pred] for pred in preds]
            except:
                warnings.warn("Failed to convert from numeric labels to text labels. Text labels are kept. ")

        assert len(preds) == df.shape[0], "DEBUG: Length mismatch"

        if isinstance(on, str):
            df[f"clf_preds_{on}"] = preds
        else:
            df[f"clf_preds_{'-'.join(on)}"] = preds

    # infer using OpenAI GPT API
    else:
        if openai_topic is None and openai_prompt is None:
            raise ValueError("Must provide either openai_topic or openai_prompt to use OpenAI classification")

        # use default label dict if it is not provided
        if label_map is None:
            label_map = {"Yes": 1, "No": 0}
            if openai_prompt is not None:
                warnings.warn("You are using a customized prompt but a default label map (Yes/No mapping). ")

        # check if label dict is compatible with num of labels
        if len(label_map) != num_labels:
            raise ValueError(f"label_dict has {label_map} entries, but num_labels is {num_labels}. ")

        # join texts in columns if needed
        if isinstance(on, list):
            strings_col = serialize_columns(df, on, sep_token=" ")
        else:
            strings_col = df[on].tolist()

        # get predictions from openai
        preds = predict_rows_with_openai(
            strings_col, model, openai_key, openai_topic, openai_prompt, openai_params, label_map
        )

        assert len(preds) == df.shape[0], "DEBUG: Length mismatch"

        if isinstance(on, str):
            df[f"clf_preds_{on}"] = preds
        else:
            df[f"clf_preds_{'-'.join(on)}"] = preds

    return df













##Add a function to test nn of each row within a df. 



###Preprocess datasets for training linker
###Prepare data for fine-tuning sbert

import pandas as pd
from sklearn.model_selection import train_test_split
from collections import defaultdict
from typing import List, Tuple, Dict, Any, Union
import pandas as pd

from linktransformer.utils import serialize_columns
from linktransformer.modified_sbert.cluster_fns import clusters_from_edges

def convert_to_text(unicode_string):
    return unicode_string.encode('ascii','ignore').decode('ascii')


def check_clust_data(data,model,text_col_names,clus_id_col_name):

    ###If *_id_name is a string, convert to list
    if isinstance(clus_id_col_name, str):
        clus_id_col_name = [clus_id_col_name]
    
    if isinstance(text_col_names, str):
        text_col_names = [text_col_names]
    
    ### Check if the columns are present in the data or not
    for col in text_col_names:
        if col not in data.columns:
            raise ValueError(f"Column {col} not present in data, please check the left column names")
    ###Check if the clus_id_col_name column is present in the data
    if clus_id_col_name:
        if clus_id_col_name[0] not in data.columns:
            raise ValueError(f"Column {clus_id_col_name} not present in data, please check the label column name")
        
    ### Drop row if all of the left or right columns are empty
    data = data.dropna(subset=text_col_names, how="all")

    # Drop if any id column is missing
    data = data.dropna(subset=clus_id_col_name, how="any")

    ### If clus_id_col_name is not empty, check if it is present in the data
    if clus_id_col_name:
        if clus_id_col_name[0] not in data.columns:
            raise ValueError(f"Column {clus_id_col_name} not present in data, please check the left id column name")
        
    ## Check if text columns form a unique key
    if not data[text_col_names].duplicated().any():
        print(f" Warning text columns do not form a unique key, please check the text column names")
    
    ###Make an idcol as group id
    data["cluster_assignment"] = data.groupby(clus_id_col_name).ngroup().astype(str) + "_g"
    clus_id_col_name="cluster_assignment"

    ## Serialize if there are more than one columns. Feature plan: Later implement deduplication and expansion to avoid rembedding the same string multiple times
    if len(text_col_names) > 1:
        print("Serializing text columns")
        data["text"] = serialize_columns(data, text_col_names, model=model)
    else:
        data["text"] = data[text_col_names[0]]
    return data, clus_id_col_name

    


def check_and_prep_data(data, model,left_col_names, right_col_names, left_id_name, right_id_name, label_col_name):

    ###If *_id_name is a string, convert to list
    if isinstance(left_id_name, str):
        left_id_name = [left_id_name]
    if isinstance(right_id_name, str):
        right_id_name = [right_id_name]

    ### Check if the columns are present in the data or not
    for col in left_col_names:
        if col not in data.columns:
            raise ValueError(f"Column {col} not present in data, please check the left column names")
    
    for col in right_col_names:
        if col not in data.columns:
            raise ValueError(f"Column {col} not present in data, please check the right column names")
        
    ###Check if the label column is present in the data
    if label_col_name:
        if label_col_name not in data.columns:
            raise ValueError(f"Column {label_col_name} not present in data, please check the label column name")
    
    
    ### Drop row if all of the left or right columns are empty
    if  left_col_names:
        data = data.dropna(subset=left_col_names, how="all")
    if  right_col_names:
        data = data.dropna(subset=right_col_names, how="all")

    # Drop if any id column is missing
    if left_id_name:
        data = data.dropna(subset=left_id_name, how="any")
    if right_id_name:
        data = data.dropna(subset=right_id_name, how="any")

    # Drop if any id column is missing
    if left_id_name and right_id_name:
        data = data.dropna(subset=left_id_name+right_id_name, how="any")

    ### If left_id_names is not empty, check if it is present in the data
    if left_id_name:
        if left_id_name[0] not in data.columns:
            raise ValueError(f"Column {left_id_name} not present in data, please check the left id column name")
    
    ### If right_id_name is not empty, check if it is present in the data
    if right_id_name:
        if right_id_name[0] not in data.columns:
            raise ValueError(f"Column {right_id_name} not present in data, please check the right id column name")
    
    ## Check if left columns form a unique key
    if not data[left_col_names].duplicated().any():
        print(f" Warning Left columns do not form a unique key, please check the left column names")
    
    #### Check if right columns form a unique key, leave a warning if not and drop duplicates
    if data[right_col_names].duplicated().any():
        print("Warning: Right columns do not form a unique key, dropping duplicates. Matching will proceed")
        data = data.drop_duplicates(subset=right_col_names)

    ## If left_id_name is not specified, we will assume left data is unique and use the index as the id
    if not left_id_name:
        data["left_id"] =  data.groupby(left_col_names).ngroup().astype(str) + "_l"
        left_id_name = "left_id"
        print("Warning: left id column not specified, using the left columns to group the data and using the groupby index as the id")
    else:
        ### If left id is specified, we can group by the left id column and use the groupby index as the id
        data["left_id"] = data.groupby(left_id_name).ngroup().astype(str) + "_l" ## Convert to string
        left_id_name = "left_id"
    
    ### If right id is not specified, we can group by the right columns and use the groupby index as the id
    if not right_id_name:
        data["right_id"] = data.groupby(right_col_names).ngroup().astype(str) + "_r"
        right_id_name = "right_id"
        print("Warning: right id column not specified, using the right columns to group the data and using the groupby index as the id")

    else:
        ## If right id is specified, we can group by the right id column and use the groupby index as the id
        data["right_id"] = data.groupby(right_id_name).ngroup().astype(str) + "_r"
        right_id_name = "right_id"

    
    ## Serialize if there are more than one columns. Feature plan: Later implement deduplication and expansion to avoid rembedding the same string multiple times
    if len(left_col_names) > 1:
        print("Serializing left columns")
        data["left_text"] = serialize_columns(data, left_col_names, model=model)
    else:
        data["left_text"] = data[left_col_names[0]]
    
    if len(right_col_names) > 1:
        data["right_text"] = serialize_columns(data, right_col_names, model=model)
    else:
        data["right_text"] = data[right_col_names[0]]
    return data, left_id_name, right_id_name



def preprocess_any_data(data: Union[str, pd.DataFrame]=None,
                        train_data: Union[str, pd.DataFrame] = None,
                        val_data: Union[str, pd.DataFrame] = None,
                        test_data: Union[str, pd.DataFrame] = None,
        left_col_names: List[str] = None,
        right_col_names: List[str] = None,
        left_id_name: List[str] = None,
        right_id_name: List[str] = None,
        label_col_name: str = None,
        clus_id_col_name: str = None,
        clus_text_col_names: List[str] = None,
        model: str = "sentence-transformers/all-mpnet-base-v2",
        val_perc: float = 0.2,
        val_query_prop:float =0.5,
        large_val: bool = True,
        test_at_end=True
                        ):
    """
    Wrapper around the two preprocessing functions. if label_col_name is not none, use prep_paired_label_data; otherwise use prep_linkage_data

    :param: data : Path to the csv or excel file or dataframe
    :param: left_col_names: List of column names to use for linkage in the left dataframe. Each list is a set of columns that are used to make a pair. Left columns need to be unique.
    :param: right_col_names: List of column names to use for linkage in the right dataframe. Each list is a set of columns that are used to make a pair. Right columns can have duplicates.
    :param: left_id_name: List of column names to use as id for the left columns.
    :param: right_id_name: List of column names to use as id for the right columns.
    :param label_col_name: Name of the column to use as label (0 or 1). Specify if you have a label column
    :param clusterid_col_name: Name of the column to use as cluster id. Specify if you have a cluster id column
    :param cluster_text_col_name: Name of the column to use as text for the cluster. Specify if you have a cluster text column
    :param: model: Model to use for tokenization.
    :param: val_perc: Percentage of data to use for validation. Defaults to 0.2.
    :param: large_val: If True, use the training data as part of the validation data (in the corpus for information retrieval evaluation)
    :return: A tuple containing the training data dictionary and the validation data, test tuple.
    """
    #Validate inputs
    if label_col_name and clus_id_col_name:
        raise ValueError("Please specify either label_col_name or clusterid_col_name. Not both")
    elif label_col_name and clus_text_col_names:
        raise ValueError("Please specify either label_col_name or cluster_text_col_name. Not both")
    elif clus_id_col_name and not clus_text_col_names:
        raise ValueError("Please specify cluster_text_col_name if you specify clusterid_col_name")
    elif clus_text_col_names and not clus_id_col_name:
        raise ValueError("Please specify clusterid_col_name if you specify cluster_text_col_name")
        

    if label_col_name:
        return prep_paired_label_data(
            data = data,
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            left_col_names=left_col_names,
            right_col_names=right_col_names,
            left_id_name=left_id_name,
            right_id_name=right_id_name,
            label_col_name=label_col_name,
            model=model,
            val_perc=val_perc,
            test_at_end=test_at_end
        )
    elif clus_id_col_name and clus_text_col_names:
        return prep_clus_data(
            data = data,
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            text_col_names=clus_text_col_names,
            clus_id_col_name=clus_id_col_name,
            model=model,
            val_perc=val_perc,
            val_query_prop=val_query_prop,
            test_at_end=test_at_end,
            large_val=large_val
        )


    else:
        return prep_linkage_data(
            data = data,
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            left_col_names=left_col_names,
            right_col_names=right_col_names,
            left_id_name=left_id_name,
            right_id_name=right_id_name,
            model=model,
            val_perc=val_perc,
            large_val=large_val,
            test_at_end=test_at_end
        )
    
    


def prep_paired_label_data(
        data = None,
        train_data=None,
        val_data=None,
        test_data=None,
        left_col_names: List[str] = [],
        right_col_names: List[str] = [],
        left_id_name: List[str] = [],
        right_id_name: List[str] = [],
        label_col_name: str = "label",
        model: str = "sentence-transformers/all-mpnet-base-v2",
        val_perc: float = 0.2,
        test_at_end=True
) ->  Tuple[Dict[str, List[str]], Tuple[List[str], List[str], List[str]],Tuple[List[str], List[str], List[str]]]:
    """
    This method is used to prepare training data + evaluation data. Evaluation is suitable for classification evaluation.
    What it would give for the eval data is a tuple of sentences1, sentences2, labels. 

    :param: data : Path to the csv or excel file or dataframe
    :param: train_data: Path to the csv or excel file or dataframe for training data
    :param: val_data: Path to the csv or excel file or dataframe for validation data
    :param: test_data: Path to the csv or excel file or dataframe for test data
    :param: left_col_names: List of column names to use for linkage in the left dataframe. Each list is a set of columns that are used to make a pair. Left columns need to be unique.
    :param: right_col_names: List of column names to use for linkage in the right dataframe. Each list is a set of columns that are used to make a pair. Right columns can have duplicates.
    :param: left_id_name: List of column names to use as id for the left columns.
    :param: right_id_name: List of column names to use as id for the right columns.
    :param label_col_name: Name of the column to use as label (0 or 1)
    :param: model: Model to use for tokenization.
    :param: val_perc: Percentage of data to use for validation. Defaults to 0.2.
    :return: A tuple containing the training data dictionary and the validation data, test tuple.
    """

    ###Check that only data or all of train, val, test are specified
    if data is not None and (train_data is not None or val_data is not None or test_data is not None):
        raise ValueError("Please specify either data or train, val, test data. Not both")
    elif data is  None and  (train_data is  None and val_data is  None and test_data is None):
        raise ValueError("Please specify either data or train, val, test data. Not none")
    elif data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(data, pd.DataFrame):
            data = data
        elif data.endswith(".xlsx"):
            data = pd.read_excel(data)
        elif data.endswith(".csv"):
            data = pd.read_csv(data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
    ##Reset indices of data
        data=data.reset_index(drop=True)

    elif train_data is not None and val_data is not None and test_data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(train_data, pd.DataFrame):
            train_data = train_data
        elif train_data.endswith(".xlsx"):
            train_data = pd.read_excel(train_data)
        elif train_data.endswith(".csv"):
            train_data = pd.read_csv(train_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(val_data, pd.DataFrame):
            val_data = val_data
        elif val_data.endswith(".xlsx"):
            val_data = pd.read_excel(val_data)
        elif val_data.endswith(".csv"):
            val_data = pd.read_csv(val_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(test_data, pd.DataFrame):
            test_data = test_data
        elif test_data.endswith(".xlsx"):
            test_data = pd.read_excel(test_data)
        elif test_data.endswith(".csv"):
            test_data = pd.read_csv(test_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
        
        ##Reset indices of train, val, test data
        train_data=train_data.reset_index(drop=True)
        val_data=val_data.reset_index(drop=True)
        test_data=test_data.reset_index(drop=True)
        
        data=train_data


    data = data.copy()

    data, left_id_rename, right_id_rename = check_and_prep_data(data, model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name)
    ### We now want to make a network from this data. Simply use the left_id and right_id columns as edges.
    ### Once we do that, we want to make connected components from this network. Each connected component is a cluster that is used as a "class" for training.
    edge_list = list(zip(data[left_id_rename], data[right_id_rename]))
    cluster_assignment = clusters_from_edges(edge_list)
    ## Assign cluster ids to each node
    ## Now, make a mapping between left_id and cluster id - reverse the dict
    cluster_assignment = {k: v for v, l in cluster_assignment.items() for k in l}
    data["cluster_assignment"] = data[left_id_rename].map(cluster_assignment)

    ### We want to split the data by cluster assignment 
    if val_perc == 1:
        train_data = data
        val_data = data
        ## Throw warning that train=val data
        print("Warning: train and val data are the same")
    else:
        if  val_data is None  and  test_data is None:
        ### Split by val perc - we want to split by cluster assignment. Only 20% of the clusters should be in the val set
            train_cluster_assignment, val_cluster_assignment = train_test_split(list(set(data["cluster_assignment"])), test_size=val_perc, random_state=42)
            ###Split val into test and val
            if test_at_end:
                print("Splitting val into test and val (equally) ")
                val_cluster_assignment, test_cluster_assignment = train_test_split(val_cluster_assignment, test_size=0.5, random_state=42)
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
                test_data = data[data["cluster_assignment"].isin(test_cluster_assignment)]
            else:
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
            
        else:
            val_data,left_id_rename, right_id_rename = check_and_prep_data(val_data,model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name)
            test_data,left_id_rename, right_id_rename = check_and_prep_data(test_data, model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name)
            
    train_data=data
    ### Now, group by cluster assignment and make a dict with cluster_assignment:[left_text, right_text1, right_text2, right_text3...]
    train_data_dict = defaultdict(list)
    for index, row in train_data.iterrows():
        train_data_dict[row["cluster_assignment"]].append(row["left_text"])
        train_data_dict[row["cluster_assignment"]].append(row["right_text"])          
    
    ### Deduplicate the lists
    train_data_dict = {k: list(set(v)) for k, v in train_data_dict.items()}

    ###Time for validation set now.
    ###Simply make a list of left_text, right_text1, right_text2, right_text3... and make a list of labels
    val_left_text_list = []
    val_right_text_list = []
    val_labels_list = []

    for index, row in val_data.iterrows():
        val_left_text_list.append(row["left_text"])
        val_right_text_list.append(row["right_text"])
        val_labels_list.append(row[label_col_name])

    ###Time for test set now.
    if test_at_end:
    ###Simply make a list of left_text, right_text1, right_text2, right_text3... and make a list of labels
        test_left_text_list = []
        test_right_text_list = []
        test_labels_list = []

        for index, row in test_data.iterrows():
            test_left_text_list.append(row["left_text"])
            test_right_text_list.append(row["right_text"])
            test_labels_list.append(row[label_col_name])
    else:
        test_left_text_list, test_right_text_list, test_labels_list=None,None,None

    return train_data_dict, (val_left_text_list, val_right_text_list, val_labels_list), (test_left_text_list, test_right_text_list, test_labels_list)    



def prep_linkage_data(
    data: Union[str, pd.DataFrame] = None,
    train_data: Union[str, pd.DataFrame] =None,
    val_data: Union[str, pd.DataFrame] =None,
    test_data: Union[str, pd.DataFrame] =None,
    left_col_names: List[str] = [],
    right_col_names: List[str] = [],
    left_id_name: List[str] = [],
    right_id_name: List[str] = [],
    model: str = "sentence-transformers/all-mpnet-base-v2",
    val_perc: float = 0.2,
    large_val: bool = True,
    test_at_end=True
) -> Tuple[Dict[str, List[str]], Tuple[Dict[str, str], Dict[str, str], Dict[str, set]],Tuple[Dict[str, str], Dict[str, str], Dict[str, set]]] :
    """
    Process linkage data to create training and validation sets. In this method, we get validation data suitable for information retrieval evaluation.
    What that means is that things queries are looked for in a corpus - an information retrieval task. 

    Args:
        data (str or df): Path to the csv or excel file or dataframe
        train_data (str or df): Path to the csv or excel file or dataframe for training data
        val_data (str or df): Path to the csv or excel file or dataframe for validation data
        test_data (str or df): Path to the csv or excel file or dataframe for test data
        left_col_names (List[str]): List of column names to use for linkage in the left dataframe. Each list is a set of columns that are used to make a pair. Left columns need to be unique.
        right_col_names (List[str]): List of column names to use for linkage in the right dataframe. Each list is a set of columns that are used to make a pair. Right columns can have duplicates.
        left_id_name (List[str]): List of column names to use as id for the left columns.
        right_id_name (List[str]): List of column names to use as id for the right columns.
        model (str): Model to use for tokenization.
        val_perc (float): Percentage of data to use for validation. Defaults to 0.2.
        large_val (bool): If True, use the training data as part of the validation data.

    Returns:
        Tuple[Dict[str, List[str]], Tuple[Dict[str, str], Dict[str, str], Dict[str, set]]]: A tuple containing the training data dictionary and the validation data tuple.

    """


    ###Check that only data or all of train, val, test are specified
    if data is not None and (train_data is not None or val_data is not None or test_data is not None):
        raise ValueError("Please specify either data or train, val, test data. Not both")
    elif data is  None and  (train_data is  None and val_data is  None and test_data is None):
        raise ValueError("Please specify either data or train, val, test data. Not none")
    elif data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(data, pd.DataFrame):
            data = data
        elif data.endswith(".xlsx"):
            data = pd.read_excel(data)
        elif data.endswith(".csv"):
            data = pd.read_csv(data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
    ##Reset indices of data
        data=data.reset_index(drop=True)

    elif train_data is not None and val_data is not None and test_data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(train_data, pd.DataFrame):
            train_data = train_data
        elif train_data.endswith(".xlsx"):
            train_data = pd.read_excel(train_data)
        elif train_data.endswith(".csv"):
            train_data = pd.read_csv(train_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(val_data, pd.DataFrame):
            val_data = val_data
        elif val_data.endswith(".xlsx"):
            val_data = pd.read_excel(val_data)
        elif val_data.endswith(".csv"):
            val_data = pd.read_csv(val_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(test_data, pd.DataFrame):
            test_data = test_data
        elif test_data.endswith(".xlsx"):
            test_data = pd.read_excel(test_data)
        elif test_data.endswith(".csv"):
            test_data = pd.read_csv(test_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
        
        ##Reset indices of train, val, test data
        train_data=train_data.reset_index(drop=True)
        val_data=val_data.reset_index(drop=True)
        test_data=test_data.reset_index(drop=True)
        
        data=train_data

    data = data.copy()

    data, left_id_rename, right_id_rename = check_and_prep_data(data, model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name=None)
    
    ### We now want to make a network from this data. Simply use the left_id and right_id columns as edges.
    ### Once we do that, we want to make connected components from this network. Each connected component is a cluster that is used as a "class" for training.
    edge_list = list(zip(data[left_id_rename], data[right_id_rename]))
    cluster_assignment = clusters_from_edges(edge_list)
    ## Assign cluster ids to each node
    ## Now, make a mapping between left_id and cluster id - reverse the dict
    cluster_assignment = {k: v for v, l in cluster_assignment.items() for k in l}
    data["cluster_assignment"] = data[left_id_rename].map(cluster_assignment)


    train_data=data
    ### We want to split the data by cluster assignment 
    if val_perc == 1:
        train_data = data
        val_data = data
        ## Throw warning that train=val data
        print("Warning: train and val data are the same")
    else:
        if  val_data is None  and  test_data is None:
            ### Split by val perc - we want to split by cluster assignment. Only 20% of the clusters should be in the val set
            train_cluster_assignment, val_cluster_assignment = train_test_split(list(set(data["cluster_assignment"])), test_size=val_perc, random_state=42)
            
            ###Split val into test and val
            if test_at_end:
                print("Splitting val into test and val (equally) ")
                val_cluster_assignment, test_cluster_assignment = train_test_split(val_cluster_assignment, test_size=0.5, random_state=42)
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
                test_data = data[data["cluster_assignment"].isin(test_cluster_assignment)]
            else:
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
            
        else:
            val_data,left_id_rename, right_id_rename = check_and_prep_data(val_data,model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name)
            
            if test_data is not None:
                test_data,left_id_rename, right_id_rename = check_and_prep_data(test_data, model, left_col_names, right_col_names, left_id_name, right_id_name, label_col_name)
            

    ### Now, group by cluster assignment and make a dict with cluster_assignment:[left_text, right_text1, right_text2, right_text3...]
    train_data_dict = defaultdict(list)
    for index, row in train_data.iterrows():
        train_data_dict[row["cluster_assignment"]].append(row["left_text"])
        train_data_dict[row["cluster_assignment"]].append(row["right_text"])          
    
    ### Deduplicate the lists
    train_data_dict = {k: list(set(v)) for k, v in train_data_dict.items()}

    queries = {}
    corpus = {}
    relevant_docs = defaultdict(set)

    for index, row in val_data.iterrows():
        queries[row[left_id_rename]] = row["left_text"]
        corpus[row[right_id_rename]] = row["right_text"]
        relevant_docs[row[left_id_rename]].add(row[right_id_rename])

    ## If large_val is True, add the training data to the validation data corpus
    if large_val:
        ##First, need to rename the 
        for index, row in train_data.iterrows():
            corpus[row[right_id_rename]+"_train"] = row["right_text"]
            relevant_docs[row[left_id_rename]+"_train"].add(row[right_id_rename]+"_train")

    val_data = queries, corpus, relevant_docs ## Needed for the evaluation function

    ##Now, prepare the test data
    if test_at_end:
        queries = {}
        corpus = {}
        relevant_docs = defaultdict(set)

        for index, row in test_data.iterrows():
            queries[row[left_id_rename]] = row["left_text"]
            corpus[row[right_id_rename]] = row["right_text"]
            relevant_docs[row[left_id_rename]].add(row[right_id_rename])
        
        test_data = queries, corpus, relevant_docs
    else:
        test_data = None


    return train_data_dict, val_data, test_data
    

def preprocess_mexican_tarrif_data(file_path="/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/wiki_data/es_mexican_products.xlsx"):
    ##Load the df
    df=pd.read_excel(file_path)
    df=df.copy()
    ##Drop missing descriptions
    df=df.dropna(subset=["description47","description48"])
    ##Make descriptions lowercase
    df=df.applymap(lambda x: x.lower() if isinstance(x,str) else x)
    ##Drop dropna rows
    df=df.dropna()
    ##Get the left and right ids
    df1=df[["tariffcode47","description47"]]
    ###Add a ground truth column
    df1["ground_truth"]=df["description48"]
    ##Print duplcated rows
    df2=df[["tariffcode48","description48"]]
    ##Drop duplicates from df1
    df1=df1.drop_duplicates(subset=["tariffcode47","description47"])
    df1=df1.drop_duplicates(subset=["description47"])

    df1.to_csv("df1.csv",index=False)
    df2.to_csv("df2.csv",index=False)

    return df1,df2


def prep_clus_data(
    data: Union[str, pd.DataFrame] = None,
    train_data: Union[str, pd.DataFrame] =None,
    val_data: Union[str, pd.DataFrame] =None,
    test_data: Union[str, pd.DataFrame] =None,
    text_col_names: List[str] = None,
    clus_id_col_name: List[str] = None,
    model: str = "sentence-transformers/all-mpnet-base-v2",
    val_perc: float = 0.2,
    val_query_prop: float = 0.5,
    large_val: bool = True,
    test_at_end=True
) -> Tuple[Dict[str, List[str]], Tuple[Dict[str, str], Dict[str, str], Dict[str, set]],Tuple[Dict[str, str], Dict[str, str], Dict[str, set]]] :
    """

    """
    
    ###Check that only data or all of train, val, test are specified
    if data is not None and (train_data is not None or val_data is not None or test_data is not None):
        raise ValueError("Please specify either data or train, val, test data. Not both")
    elif data is  None and  (train_data is  None and val_data is  None and test_data is None):
        raise ValueError("Please specify either data or train, val, test data. Not none")
    elif data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(data, pd.DataFrame):
            data = data
        elif data.endswith(".xlsx"):
            data = pd.read_excel(data)
        elif data.endswith(".csv"):
            data = pd.read_csv(data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
    ##Reset indices of data
        data=data.reset_index(drop=True)

    elif train_data is not None and val_data is not None and test_data is not None:
        ### Load the data if xlsx, else csv
        if isinstance(train_data, pd.DataFrame):
            train_data = train_data
        elif train_data.endswith(".xlsx"):
            train_data = pd.read_excel(train_data)
        elif train_data.endswith(".csv"):
            train_data = pd.read_csv(train_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(val_data, pd.DataFrame):
            val_data = val_data
        elif val_data.endswith(".xlsx"):
            val_data = pd.read_excel(val_data)
        elif val_data.endswith(".csv"):
            val_data = pd.read_csv(val_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")

        if isinstance(test_data, pd.DataFrame):
            test_data = test_data
        elif test_data.endswith(".xlsx"):
            test_data = pd.read_excel(test_data)
        elif test_data.endswith(".csv"):
            test_data = pd.read_csv(test_data)
        else:
            raise ValueError("Data should be a path to a csv or excel file or a dataframe")
        
        ##Reset indices of train, val, test data
        train_data=train_data.reset_index(drop=True)
        val_data=val_data.reset_index(drop=True)
        test_data=test_data.reset_index(drop=True)
        
        data=train_data

    data = data.copy()

    data,cluster_id_rename=check_clust_data(data, model, text_col_names, clus_id_col_name)
    
    
    train_data=data
    ### We want to split the data by cluster assignment 
    if val_perc == 1:
        train_data = data
        val_data = data
        ## Throw warning that train=val data
        print("Warning: train and val data are the same")
    else:
        if  val_data is None  and  test_data is None:
            ### Split by val perc - we want to split by cluster assignment. Only 20% of the clusters should be in the val set
            train_cluster_assignment, val_cluster_assignment = train_test_split(list(set(data["cluster_assignment"])), test_size=val_perc, random_state=42)
            
            ###Split val into test and val
            if test_at_end:
                print("Splitting val into test and val (equally) ")
                val_cluster_assignment, test_cluster_assignment = train_test_split(val_cluster_assignment, test_size=0.5, random_state=42)
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
                test_data = data[data["cluster_assignment"].isin(test_cluster_assignment)]
            else:
                train_data = data[data["cluster_assignment"].isin(train_cluster_assignment)]
                val_data = data[data["cluster_assignment"].isin(val_cluster_assignment)]
            
        else:
            val_data,cluster_id_rename  = check_clust_data(val_data, model, text_col_names, clus_id_col_name)
    
            if test_data is not None:
                test_data,cluster_id_rename = check_clust_data(test_data, model, text_col_names, clus_id_col_name)
    

    ### Now, group by cluster assignment and make a dict with cluster_assignment:[left_text, right_text1, right_text2, right_text3...]
    # print("Preparing train data")
    # train_data_dict = defaultdict(list)
    # for index, row in train_data.iterrows():
    #     train_data_dict[row["cluster_assignment"]].append(row["text"])
    
    # ### Deduplicate the lists
    # train_data_dict = {k: list(set(v)) for k, v in train_data_dict.items()}
    train_data_dict = train_data.groupby("cluster_assignment")["text"].apply(set).apply(list).to_dict()

    print("Preparing val data")

    queries = {}
    corpus = {}
    relevant_docs = defaultdict(set)

    ###We want 10% of the cluster to form queries, the rest corpus. 
    ###We should iterate over groups for this
    for cluster_id, group in val_data.groupby(cluster_id_rename):
        query_portion=int(len(group)*val_query_prop)
        query_group=group.iloc[:query_portion].reset_index(drop=True)
        corpus_group=group.iloc[query_portion:].reset_index(drop=True)
        query_group_ids=query_group[cluster_id_rename].tolist()
        query_group_ids=[query_group_ids[i]+str(i) for i in range(len(query_group_ids))]

        corpus_group_ids=corpus_group[cluster_id_rename].tolist()
        corpus_group_ids=[corpus_group_ids[i]+str(i)+"c" for i in range(len(corpus_group_ids))]

        ##Relevant docs contain all corpus ids for each query
        for qindex, query_id in enumerate(query_group_ids):
            queries[query_id]=query_group["text"].iloc[qindex]
            relevant_docs[query_id]=set(corpus_group_ids)

        for cindex,corpus_id in enumerate(corpus_group_ids):
            corpus[corpus_id]=corpus_group["text"].iloc[cindex]




    if large_val:
        ##ADd train data to the corpus 
        for cluster_id, group in train_data.groupby(cluster_id_rename):
            group=group.reset_index(drop=True)
            corpus_group_ids=group[cluster_id_rename].tolist()
            corpus_group_ids=[corpus_group_ids[i]+str(i)+"ct" for i in range(len(corpus_group_ids))]


            for cindex,corpus_id in enumerate(corpus_group_ids):
                corpus[corpus_id]=group["text"].iloc[cindex]

    val_data = queries, corpus, relevant_docs ## Needed for the evaluation function

    ##Now, prepare the test data
    if test_at_end:
        print("Preparing test data")

        queries = {}
        corpus = {}
        relevant_docs = defaultdict(set)
        for cluster_id, group in test_data.groupby(cluster_id_rename):
            query_portion=int(len(group)*val_query_prop)
            query_group=group.iloc[:query_portion]
            corpus_group=group.iloc[query_portion:]
            query_group_ids=query_group[cluster_id_rename].tolist()
            query_group_ids=[query_group_ids[i]+str(i) for i in range(len(query_group_ids))]
            ###Add an index after each query_group_id
            # for index,group_id in enumerate(query_group_ids) :
            #     query_group_ids[index]=query_group_ids[index]+"_"+str(index)
            
            corpus_group_ids=corpus_group[cluster_id_rename].tolist()
            corpus_group_ids=[corpus_group_ids[i]+str(i)+"c" for i in range(len(corpus_group_ids))]

            ##Relevant docs contain all corpus ids for each query
            for qindex, query_id in enumerate(query_group_ids):
                queries[query_id]=query_group["text"].iloc[qindex]
                relevant_docs[query_id]=set(corpus_group_ids)

            for cindex,corpus_id in enumerate(corpus_group_ids):
                corpus[corpus_id]=corpus_group["text"].iloc[cindex]
            
        test_data = queries, corpus, relevant_docs
    else:
        test_data = None


    return train_data_dict, val_data, test_data
    






##Run as script
if __name__ == "__main__":
   
    # train_data,val_data=prep_linkage_data("/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/wiki_data/es_mexican_products.xlsx",left_col_names=["description47"],right_col_names=['description48'],left_id_name=['tariffcode47'],right_id_name=['tariffcode48'],model="all-mpnet-base-v2",val_perc=0.2,large_val=False)

    # train_data,val_data,test_data=prep_clus_data(data="/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/linktransformer/src/linktransformer/data/company_clusters.csv",text_col_names=["company_name"],clus_id_col_name=["cluster_id"],model="all-mpnet-base-v2",val_perc=0.2,large_val=False)
    # # train_data,val_data,test_data=preprocess_any_data(data="/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/linktransformer/src/linktransformer/data/company_clusters.csv",text_col_names=["company_name"],clus_id_col_name=["cluster_id"],model="all-mpnet-base-v2",val_perc=0.2,large_val=False)
    # print(train_data)
    # print(val_data)
    # print(test_data)
    # train_data,val_data,test_data=preprocess_any_data(data="/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/linktransformer/src/linktransformer/data/company_clusters.csv",clus_text_col_names=["company_name"],clus_id_col_name=["cluster_id"],model="all-mpnet-base-v2",val_perc=0.2,large_val=True)
    # print(train_data)
    # print(val_data)
    # print(test_data)
    train_data,val_data,test_data=preprocess_any_data(data="/mnt/122a7683-fa4b-45dd-9f13-b18cc4f4a187/deeprecordlinkage/linktransformer/src/linktransformer/data/es_mexican_products.xlsx",
                                                      left_col_names=["description47"],
                                                        right_col_names=['description48'],
                                                        left_id_name=['tariffcode47'],
                                                        right_id_name=['tariffcode48'],
                                                      model="all-mpnet-base-v2",
                                                      val_perc=0.2,large_val=True)
    print(train_data)
    print(val_data)
    print(test_data)


import os
import time
import warnings
from typing import List, Optional

from linktransformer.modelling.LinkTransformer import LinkTransformer
import numpy as np
import pandas as pd
import transformers
import openai
from datasets import Dataset
from itertools import combinations

from tqdm import tqdm


def load_model(model_path: str) -> LinkTransformer:
    """
    Load a saved LinkTransformer model.

    :param model_path: Path to the saved model.
    :return: The loaded LinkTransformer model.
    """
    ###Throw an error if there is an error in loading the model
    try:
        model = LinkTransformer(model_path)
    except:
        print("Can't load the model, please check your path. All transformer based models from [HuggingFace](https://huggingface.co/) are supported. \
               We recommend [sentence-transformers](https://www.sbert.net/docs/pretrained_models.html) for these tasks as they are trained for \
               semantic similarity tasks. ")

    return model


def load_clf(model_path: str, num_labels: int = 2) -> transformers.AutoModelForSequenceClassification:
    """
    Load the classification model.

    :param model_path: Path to the saved model.
    :param num_labels: Number of classes to predict (default is two)
    :return: The loaded Classification model
    """
    try:
        model = transformers.AutoModelForSequenceClassification.from_pretrained(model_path, num_labels=num_labels)
    except Exception as e:
        print(repr(e))
        raise ValueError("Error loading classification model. ")

    return model


def cosine_similarity_corresponding_pairs(vector1, vector2):
    dot_product = np.sum(vector1 * vector2, axis=1)
    norm_vector1 = np.linalg.norm(vector1, axis=1)
    norm_vector2 = np.linalg.norm(vector2, axis=1)
    cosine_sim = dot_product / (norm_vector1 * norm_vector2)
    return cosine_sim


def serialize_columns(df: pd.DataFrame, columns: list, sep_token: str = None, model: str = None) -> list:
    """
    Serialize columns of a DataFrame into a single string.

    :param df: The DataFrame.
    :param columns: The columns to serialize.
    :param sep_token: The token to use to separate columns.
    :param model: The language model to use for tokenization (optional).
    :return: List of serialized strings.
    """
    if model is not None:
        if "/" not in model:
            print("No base organization specified, if there is an error, it is likely because of that.")
            print(
                f"Trying to append the model : sentence-transformers/{model} and linktransformers/{model}. Check your path otherwise!")
            ###Error handling
            try:
                tokenizer = transformers.AutoTokenizer.from_pretrained(model)
                sep_token = tokenizer.sep_token
            except:
                try:
                    print(f"Trying sentence-transformers/{model}...")
                    tokenizer = transformers.AutoTokenizer.from_pretrained("sentence-transformers/" + model)
                    sep_token = tokenizer.sep_token
                except:
                    print(f"Trying linktransformers/{model}...")
                    tokenizer = transformers.AutoTokenizer.from_pretrained("linktransformers/" + model)
                    sep_token = tokenizer.sep_token
        else:
            tokenizer = transformers.AutoTokenizer.from_pretrained(model)
            sep_token = tokenizer.sep_token

    return df[columns].apply(lambda x: sep_token.join(x.astype(str)), axis=1).tolist()


def infer_embeddings(strings: list, model: LinkTransformer, batch_size: int = 128,
                     sbert: bool = True, openai_key: str = None) -> np.ndarray:
    """
    Infer embeddings for a list of strings using a language model.

    :param strings: List of strings.
    :param model: The language model to use for inference.
    :param batch_size: Batch size for inference (default: 128).
    :param sbert: If True, load model as LinkTransformer (default: True).
    :param openai_key: OpenAI API key (optional).
    :return: Embeddings as a numpy array.
    """

    if openai_key is None:
        if isinstance(model, LinkTransformer):
            embeddings = model.encode(strings, batch_size=batch_size)
        elif isinstance(model, str):
            if sbert:
                model = LinkTransformer(model)
            else:
                model = transformers.AutoModel.from_pretrained(model)
            embeddings = model.encode(strings, batch_size=batch_size)
        else:
            raise ValueError(f"Invalid model type: {type(model)}")
    else:
        openai.api_key = openai_key
        ###Open ai has a token limit on  max number of tokens per request
        char_count_string = [len(x) for x in strings]
        ##Based on character count, split the list into multiple lists - each with a max of 5000 characters
        ##This is a very rough approximation - we can do better
        ##But this is a good starting point
        ##Get the indices of the list where the split should happen. Aggregate the character count list and split when the sum is greater than 5000
        split_indices = [0]
        char_count_sum = 0
        for i in range(len(char_count_string)):
            char_count_sum += char_count_string[i]
            if char_count_sum > 5000:
                split_indices.append(i)
                char_count_sum = 0
        split_indices.append(len(char_count_string))
        ##Split the list of strings into multiple lists
        split_strings = [strings[split_indices[i]:split_indices[i + 1]] for i in range(len(split_indices) - 1)]
        ##Get the embeddings for each of the split lists
        embeddings = []
        for i in range(len(split_strings)):
            response = openai.Embedding.create(input=split_strings[i], model=model)["data"]
            f = lambda x: x["embedding"]
            embeddings.append(np.array(list(map(f, response)), dtype=np.float32))
        embeddings = np.concatenate(embeddings, axis=0)

    return embeddings


def tokenize_data_for_inference(corpus: str, name: str, hf_model: str):
    dataset = Dataset.from_dict({name: corpus})

    # Instantiate tokenizer
    tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model)

    # Tokenize datasets
    def tokenize_function(dataset):
        return tokenizer(dataset[name], padding="max_length", truncation=True)

    tokenized_dataset = dataset.map(tokenize_function, batched=True)

    return tokenized_dataset


def get_completion_from_messages(text, model, openai_topic=None, openai_prompt=None, openai_params=None):
    """
    This function takes text of an article and send it to OpenAI API as user input and
    collects the content of the API response and the total number of tokens used

    :param text: (str) user input to API
    :param model: (str) name of the model to use (see "https://platform.openai.com/docs/models")
    :returns: tuple ((str, int)): response text from API and number of tokens used
    """

    if openai_prompt is None:
        sys_prompt = f"Determine whether the text is about {openai_topic} or not. Yes/No: "
    else:
        sys_prompt = openai_prompt

    response = openai.ChatCompletion.create(
        model=model,
        messages=[
            {
                "role": "system",
                "content": sys_prompt
            },
            {
                "role": "user",
                "content": text
            }
        ],
        temperature=openai_params["temperature"] if "temperature" in openai_params else 0,
        max_tokens=openai_params["max_tokens"] if "max_tokens" in openai_params else 1,
        top_p=openai_params["top_p"] if "top_p" in openai_params else 0,
        frequency_penalty=openai_params["frequency_penalty"] if "frequency_penalty" in openai_params else 0,
        presence_penalty=openai_params["presence_penalty"] if "presence_penalty" in openai_params else 0,
        request_timeout=openai_params["request_timeout"] if "request_timeout" in openai_params else 10,
    )
    return response.choices[0].message["content"], response.usage["total_tokens"]


def predict_rows_with_openai(
        strings_col: List[str],
        model: str = "gpt-3.5-turbo",
        openai_key: str = None,
        openai_topic: Optional[str] = None,
        openai_prompt: Optional[str] = None,
        openai_params: Optional[dict] = None,
        label_dict: Optional[dict] = None
):
    """
    This function takes a list of texts and run the texts through the API. The first part of the function
    specifies the rate-limiting parameters (e.g. how long to sleep when encountering rate-limiting errors).
    The second part retrieve the api response and save the output from API.

    :param strings_col: (List[str]) a list of texts to classify
    :param model: (str) name of the model to use (see "https://platform.openai.com/docs/models")
    :param openai_key: (str) OpenAI API key
    :param openai_topic: (str) topic to classify
    :param openai_prompt: (str) custom system prompt for OpenAI API
    :param openai_params: (dict) a dictionary to set custom parameters for OpenAI API (temperature, top_p, max_tokens, etc.)
    :param label_dict: (dict) a dictionary map text labels to numeric labels
    :returns: (List[int]) a list of labels from OpenAI API
    """

    openai.api_key = openai_key

    preds = []

    # set parameters
    if model == "gpt-3.5-turbo":
        max_tokens = 90000
        iteration_sleep_time = 0
        ratelimit_sleep_time = 60
    elif model == "gpt-4":
        max_tokens = 10000
        iteration_sleep_time = 1
        ratelimit_sleep_time = 60
    else:
        if "gpt-3.5" in model:
            max_tokens = 90000
        elif "gpt-4" in model:
            max_tokens = 10000
        else:
            max_tokens = 250000  # ada, baggage, curie, davinci (default limit)
        iteration_sleep_time = 0
        ratelimit_sleep_time = 60

    # get results from api
    accm_tokens = 0
    for i in tqdm(range(len(strings_col))):

        # sleep at when accumulated token is approaching rate-limiting threshold
        if accm_tokens > 0.8 * max_tokens:
            tqdm.write(f"Sleeping to avoid rate-limiting")
            time.sleep(ratelimit_sleep_time)
            accm_tokens = 0

        # send text at row i and fetch response
        try:
            r, num_tokens = get_completion_from_messages(
                strings_col[i], model, openai_topic, openai_prompt, openai_params
            )
            preds.append(r)

            accm_tokens += num_tokens
            time.sleep(iteration_sleep_time)
        except Exception as e:  # handles RateLimit or Timeout Error
            tqdm.write(repr(e))
            tqdm.write(f"Encountered error: will retry after sleeping")
            time.sleep(ratelimit_sleep_time)
            accm_tokens = 0

            r, num_tokens = get_completion_from_messages(
                strings_col[i], model, openai_topic, openai_prompt, openai_params
            )
            preds.append(r)

            accm_tokens += num_tokens

    if label_dict is None:
        label_dict = {"Yes": 1, "No": 0}

    try:
        preds_numeric = [label_dict[x] for x in preds]
        return preds_numeric
    except:
        warnings.warn(
            "Failed to convert OpenAI text labels to numeric labels. Text labels are kept. \
            You may want to modify the prompt or the label dict. ")
        return preds



import math
import os
import sys
from typing import Dict
from linktransformer.modified_sbert import losses, data_loaders, evaluation
from torch.utils.data import DataLoader# This is the same as torch.utils.data.DataLoader
import sentence_transformers
from sentence_transformers import models, LoggingHandler 
from sentence_transformers.datasets import SentenceLabelDataset
import logging
from transformers import logging as lg
import wandb    

from linktransformer.modelling.LinkTransformer import LinkTransformer

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
#grandparentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)
#sys.path.append(grandparentdir)




lg.set_verbosity_error()
logging.basicConfig(format='%(asctime)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    level=logging.INFO,
                    handlers=[LoggingHandler()])
logger = logging.getLogger(__name__)


def train_biencoder(
        train_data: dict = None,
        dev_data: dict = None,
        test_data: dict = None,
        base_model='colorfulscoop/sbert-base-ja',
        add_pooling_layer=False,
        train_batch_size=64,# The train_batch_size is an argument - you can change it for hard negatives training
        num_epochs=10,
        warm_up_perc=0.1,
        optimizer_params: Dict[str, object] = {'lr': 2e-7},
        loss_params=None,
        model_save_path="output",
        wandb_names=None,
        already_clustered_train=False,
        eval_steps_perc=0.1,
        eval_type="retrieval",
        opt_model_description=None,
        opt_model_lang=None

):

    # Logging
    if wandb_names: 
        if 'run' in wandb_names:
            wandb.init(project=wandb_names['project'], entity=wandb_names['id'], reinit=True, name=wandb_names['run'])
        else:
            wandb.init(project=wandb_names['project'], entity=wandb_names['id'], reinit=True)

        wandb.config = {
            "epochs": num_epochs,
            "batch_size": train_batch_size,
            "warm_up": warm_up_perc,
        }
        

    os.makedirs(model_save_path, exist_ok=True)

    # Base language model
    if add_pooling_layer:
        word_embedding_model = models.Transformer(base_model, max_seq_length=512)
        pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension(), pooling_mode='mean')
        model = LinkTransformer(modules=[word_embedding_model, pooling_model],opt_model_description=opt_model_description,opt_model_lang=opt_model_lang)
    elif type(base_model) == LinkTransformer:
        model = base_model
    else:
        model = LinkTransformer(base_model,opt_model_description=opt_model_description,opt_model_lang=opt_model_lang)

    if wandb_names is not None:
        wandb.watch(model)


    train_loss = losses.SupConLoss_wandb(model=model,**loss_params if loss_params is not None else {},wandb_names=wandb_names)    
        


    # Special dataset "SentenceLabelDataset" to wrap out train_set
    # It yields batches that contain at least two samples with the same label
    '''
    SentenceLabelDataset: 
    https://github.com/UKPLab/sentence-transformers/blob/master/sentence_transformers/datasets/SentenceLabelDataset.py

    '''

    # Load data as individuals
    ## train_data - You should organize it into cluster fformat
    train_samples = data_loaders.load_data_as_individuals(train_data, type="training",already_clustered=already_clustered_train)
    train_data_sampler = SentenceLabelDataset(train_samples)
    train_dataloader = DataLoader(train_data_sampler, batch_size=train_batch_size)

    if eval_type=="retrieval":
        print("Evaluating on retrieval task")
        queries,corpus,relevant_docs=dev_data
        evaluators = [evaluation.InformationRetrievalEvaluator_wandb(queries,corpus,relevant_docs,wandb_names=wandb_names,name="eval")]
        seq_evaluator = sentence_transformers.evaluation.SequentialEvaluator(evaluators, main_score_function=lambda scores: scores[-1])
        ###We also want to evaluate on the test set
        if test_data is not None:
            queries,corpus,relevant_docs=test_data
            test_evaluators = [evaluation.InformationRetrievalEvaluator_wandb(queries,corpus,relevant_docs,wandb_names=wandb_names,name="test")]
            test_evaluator = sentence_transformers.evaluation.SequentialEvaluator(test_evaluators, main_score_function=lambda scores: scores[-1])

    elif eval_type=="classification":
        print("Evaluating on pair-wise classification task")
        sentences1, sentences2, labels = dev_data
        evaluators = [evaluation.BinaryClassificationEvaluator_wandb(sentences1, sentences2, labels,wandb_names=wandb_names,name="eval")]
        seq_evaluator = sentence_transformers.evaluation.SequentialEvaluator(evaluators, main_score_function=lambda scores: scores[-1])
        ###We also want to evaluate on the test set
        if test_data is not None:
            sentences1, sentences2, labels = test_data
            test_evaluators = [evaluation.BinaryClassificationEvaluator_wandb(sentences1, sentences2, labels,wandb_names=wandb_names,name="test")]
            test_evaluator = sentence_transformers.evaluation.SequentialEvaluator(test_evaluators, main_score_function=lambda scores: scores[-1])
    else:
        raise ValueError("eval_type can only be either 'retrieval' or 'classification'")
    


    logger.info("Evaluate model without training")
    seq_evaluator(model, epoch=0, steps=0, output_path=model_save_path)

    # Train the model
    model.fit(
        train_objectives=[(train_dataloader, train_loss)],
        evaluator=seq_evaluator,
        epochs=num_epochs,
        warmup_steps=math.ceil(len(train_dataloader) * num_epochs * warm_up_perc),
        output_path=model_save_path,
        evaluation_steps= math.ceil(len(train_dataloader)*eval_steps_perc),
        checkpoint_save_steps=None,
        checkpoint_path=None,
        save_best_model=True,
        checkpoint_save_total_limit=None,
        optimizer_params = optimizer_params,
    )

    ###Get the best model path among the saved checkpoints
    ##take the last saved checkpoints = best
    ###look in the dir and get the last checkpoint (subdirs are named as numbers, take the max one)
    # best_model_path = os.path.join(model_save_path, str(max([int(x) for x in os.listdir(model_save_path) if x.isdigit()])))
    ###Evaluate on the test set using the best model
    if test_data is not None:
        logger.info("Evaluating on the test set (0.5 of val data)")
        ##Load the best model
        print("Loading best model from:", model_save_path)
        best_model=LinkTransformer(model_save_path)
        test_evaluator(best_model, epoch=0, steps=0, output_path=model_save_path)

    return model_save_path

